# SPDX-FileCopyrightText: BioPAL <biopal@esa.int>
# SPDX-License-Identifier: MIT

import os
import numpy as np
import logging
from shapely.geometry import Polygon
from osgeo.gdalconst import GA_ReadOnly
from osgeo import gdal
from equi7grid.equi7grid import Equi7Grid
from equi7grid.image2equi7grid import image2equi7grid
from scipy.signal import convolve2d

# biomassL2 processor imports
from biopal.agb.processing_AGB import (
    LookupTableAGB,
    initialize_inversion_parameters,
    get_projection_from_path,
)
from biopal.data_operations.data_operations import (
    read_and_oversample_data,
    read_and_oversample_aux_data,
    fnf_equi7_load_filter_equi7format,
    fnf_tandemx_load_filter_equi7format,
    apply_dem_flattening,
)
from biopal.utility.utility_functions import (
    Task,
    start_logging,
    choose_equi7_sampling,
    check_if_path_exists,
    check_if_geometry_auxiliaries_are_present,
    check_fnf_folder_format,
    check_cal_format,
    get_min_time_stamp_repository,
    get_raster_cal_names,
    get_foss_cal_names,
    resolution_heading_correction,
    decode_unique_acquisition_id_string,
    save_breakpoints,
    set_gdal_paths,
)
from biopal.geocoding.geocoding import (
    geocoding,
    geocoding_init,
)
from biopal.io.xml_io import (
    parse_input_file,
    write_input_file,
    parse_configuration_file,
    write_configuration_file,
    core_processing_agb,
)

from biopal.io.data_io import tiff_formatter
from biopal.screen_calibration.screen_calibration import apply_calibration_screens
from biopal.geometry.utility_geometry import compute_and_oversample_geometry_auxiliaries
from biopal.ground_cancellation.ground_cancellation import ground_cancellation

from biopal.agb.estimating_AGB import (
    transform_function,
    match_string_lists,
    sample_and_tabulate_data,
    fit_formula_to_random_subsets,
    save_human_readable_table,
    read_and_organise_3d_data,
    subset_iterable,
    map_space_variant_parameters,
    check_block_for_data_and_cal,
    compute_block_processing_order,
)


class InvalidInputError(ValueError):
    pass


class AboveGroundBiomass(Task):
    """AGB main APP Above Ground Biomass
    
    run this APP to execute the complete Above Ground Biomass (AGB) processing chain.
    
    AboveGroundBiomass is composed by two sub APPS automatically called in sequence  
    when standard launch is performed:
    StackBasedProcessingAGB -> CoreProcessingAGB

    Refer to dataset_query, StackBasedProcessingAGB and CoreProcessingAGB documentation 
    and Tutorial section for step by step run.

    Attributes
    ----------
    configuration_file : str
        path of the Configuration_File.xml file

    Methods
    -------
    run( input_file_path )
        run the AboveGroundBiomass processing
    name : str
        name of the APP

    See Also
    --------
    biopal.dataset_query.dataset_query.dataset_query: it's the APP to be called before this APP
    StackBasedProcessingAGB : it's the first of the two sub-APPs called by AboveGroundBiomass preocessor
    CoreProcessingAGB : it's  the second of the two sub-APPs called by AboveGroundBiomass processor

    Examples
    --------
    Manual AGB chain execution

    >>> from biopal.dataset_query.dataset_query import dataset_query
    >>> from biopal.agb.main_AGB import AboveGroundBiomass
    >>> dq_obj = dataset_query()
    >>> input_file_up = dq_obj.run( input_file )
    >>> chain_obj = AboveGroundBiomass( configuration_file )
    >>> chain_obj.run( input_file_up )

    - input_file: path of the BioPAL input file
    - input_file_up: same of input_file with also the "stack_based_processing" section
    - configuration_file: path of the BioPAL configuration file
    """

    def __init__(
        self, configuration_file,
    ):
        super().__init__(configuration_file)

    def _run(self, input_file):

        # Main APP #1: Stack Based Processing
        stack_based_processing_obj = StackBasedProcessingAGB(self.configuration_file)

        # Run Main APP #1: Stack Based Processing
        input_file_updated, configuration_file_updated = stack_based_processing_obj.run(input_file)

        # Main APP #2: AGB Core Processing
        agb_processing_obj = CoreProcessingAGB(configuration_file_updated)

        # Run Main APP #2: AGB Core Processing
        agb_processing_obj.run(input_file_updated)


class StackBasedProcessingAGB(Task):
    """StackBasedProcessingAGB APP 
    
    StackBasedProcessingAGB APP is the first of the two sub-APPs called by AboveGroundBiomass processor.

    It performs stack-based operations to prepare inputs for the "CoreProcessingAGB" APP. 
    It is automatically called in the default run (see Getting Started  and Tutorial documentation section),
    or can be manually launched (stand alone) following this sequence:
    dataset_query -> StackBasedProcessingAGB -> CoreProcessingAGB

    Attributes
    ----------
    configuration_file : str
        path of the Configuration_File.xml file

    Methods
    -------
    run( input_file_path )
        run the StackBasedProcessingAGB APP
    name : str
        name of the APP

    See Also
    --------
    biopal.dataset_query.dataset_query.dataset_query : it's the first APP to be called in the manual sequence
    CoreProcessingAGB : it's the core APP that follows this APP in the call sequence

    Examples
    --------
    Manual AGB chain execution

    >>> from biopal.dataset_query.dataset_query import dataset_query
    >>> from biopal.AGB.main_AGB import StackBasedProcessingAGB, CoreProcessingAGB
    >>> dq_obj = dataset_query()
    >>> input_file_up1 = dq_obj.run( input_file )
    >>> sbp_obj = StackBasedProcessingAGB( config_file )
    >>> input_file_up2, config_file_up = sbp_obj.run( input_file_up1 )
    >>> agbp_obj = CoreProcessingAGB( config_file_up )
    >>> agbp_obj.run( input_file_up2 )

    - input_file: path of the BioPAL input file
    - input_file_up1: same of input_file with also the "stack_based_processing" section   
    - input_file_up2: same of input_file_up1 with also the "core_processing_agb" section
    - config_file: path of the BioPAL configuration file
    - config_file_up: same of config_file, with paths updated      
    """

    def __init__(self, configuration_file):
        super().__init__(configuration_file)

    def _run(self, input_file):

        # parse the input file and check for the needed sections in it:
        if isinstance(input_file, str):
            check_if_path_exists(input_file, "FILE")
            input_params_obj = parse_input_file(input_file)
        else:
            input_params_obj = input_file
        if (
            input_params_obj.L2_product is None
            or input_params_obj.output_specification is None
            or input_params_obj.dataset_query is None
            or input_params_obj.stack_based_processing is None
        ):
            error_message = [
                "StackBasedProcessingAGB APP input file should contain at least"
                ' "L2_product", "output_specification", "dataset_query" and "stack_based_processing" sections'
            ]
            raise InvalidInputError(error_message)

        # parse the configuration file and check for the needed sections in it:
        if isinstance(input_file, str):
            check_if_path_exists(self.configuration_file, "FILE")
            conf_params_obj = parse_configuration_file(self.configuration_file)
        else:
            conf_params_obj = self.configuration_file
        if (
            conf_params_obj.processing_flags is None
            or conf_params_obj.ground_cancellation is None
            or conf_params_obj.estimate_agb is None
        ):
            error_message = [
                "Configuration file for  StackBasedProcessingAGB APP should contain at least"
                ' "processing_flags", "ground_cancellation" and "estimate_agb" sections'
            ]
            raise InvalidInputError(error_message)

        # Initialize the logger, of the APP has been called in a stand-alone call
        if not hasattr(logging.getLoggerClass().root.handlers[0], "baseFilename"):
            start_logging(
                os.path.dirname(input_params_obj.output_specification.output_folder),
                input_params_obj.L2_product,
                "DEBUG",
                app_name="StackBasedProcessingAGB",
            )
        logging.info("AGB stack-based-processing APP starting\n")

        # get needed parameters from input and configuration files
        geographic_boundaries = input_params_obj.stack_based_processing.geographic_boundaries
        geographic_boundaries_per_stack = input_params_obj.stack_based_processing.geographic_boundaries_per_stack

        gdal_path, _ = set_gdal_paths(conf_params_obj.gdal.gdal_path, conf_params_obj.gdal.gdal_environment_path)

        ########################## INITIAL STEPS ##############################

        ### managing output folders:
        products_folder = os.path.join(input_params_obj.output_specification.output_folder, "Products")
        if conf_params_obj.processing_flags.save_breakpoints:
            breakpoints_output_folder = os.path.join(products_folder, "breakpoints")
            logging.info("AGB: Breakpoints will be saved into: " + breakpoints_output_folder)
            os.makedirs(breakpoints_output_folder)

        temp_output_folder = os.path.join(products_folder, "temp")
        logging.info("AGB: Temporary data folder:" + temp_output_folder + "\n")

        os.makedirs(temp_output_folder)

        ### get temporal date time of the input data (get the minimum date from all the stacks)
        time_tag_mjd_initial = get_min_time_stamp_repository(
            input_params_obj.dataset_query.L1C_repository, input_params_obj.stack_based_processing.stack_composition
        )

        # read the cals:
        cal_format = check_cal_format(input_params_obj.stack_based_processing.reference_agb_folder)
        if cal_format == "GeoJSON":
            cal_fnames = get_foss_cal_names(input_params_obj.stack_based_processing.reference_agb_folder)
            flag_cal = 0
            pass
        elif cal_format == "RASTER":
            flag_cal = 1
            cal_fnames = get_raster_cal_names(input_params_obj.stack_based_processing.reference_agb_folder)

            # LUT: CAL
            lut_cal_boundaries = np.zeros((len(cal_fnames), 5))
            lut_cal_paths = []
            for idx_cal, equi7_cal_fname in enumerate(cal_fnames):

                data_driver = gdal.Open(equi7_cal_fname, GA_ReadOnly)
                # [ upper_left_easting_coord, sampling_step_east_west, 0, upper_left_northing_coord, 0, sampling_step_north_south ]
                geotransform_equi7 = data_driver.GetGeoTransform()
                east_len = data_driver.RasterXSize
                north_len = data_driver.RasterYSize

                data_driver = None

                east_min = geotransform_equi7[0]
                east_max = east_min + geotransform_equi7[1] * east_len
                north_in = geotransform_equi7[3]
                north_out = north_in + geotransform_equi7[5] * north_len

                lut_cal_boundaries[idx_cal, :] = [
                    east_min,
                    east_max,
                    min(north_in, north_out),
                    max(north_in, north_out),
                    flag_cal,
                ]
                lut_cal_paths.append(equi7_cal_fname)

                logging.info("using calibration file: " + equi7_cal_fname)

        ### initialize the equi7 sampling grid
        equi7_sampling_intermediate = choose_equi7_sampling(
            conf_params_obj.estimate_agb.intermediate_ground_averaging,
            conf_params_obj.estimate_agb.intermediate_ground_averaging / 2,
        )
        e7g_intermediate = Equi7Grid(equi7_sampling_intermediate)
        logging.info("EQUI7 Grid sampling used for intermediate products: {}".format(equi7_sampling_intermediate))

        # it is loaded if equi7, or converted to equi7 if TANDEM-X
        # it is a list containing all the loaded FNF-FTILES
        fnf_format = check_fnf_folder_format(input_params_obj.stack_based_processing.forest_mask_catalogue_folder)
        if fnf_format == "TANDEM-X":
            logging.info("Initial Forest mask is in TANDEM-X format, converting to equi7...")

            # conversion step 1: get the tiff names of current zone
            equi7_fnf_mask_fnames = fnf_tandemx_load_filter_equi7format(
                input_params_obj.stack_based_processing.forest_mask_catalogue_folder,
                e7g_intermediate,
                conf_params_obj.estimate_agb.intermediate_ground_averaging,
                temp_output_folder,
                gdal_path,
                geographic_boundaries,
                time_tag_mjd_initial,
            )

        elif fnf_format == "EQUI7":
            logging.info("Initial Forest mask reading and formatting...")

            # re format the input mask in equi7 with the output resolution:
            equi7_fnf_mask_fnames = fnf_equi7_load_filter_equi7format(
                input_params_obj.stack_based_processing.forest_mask_catalogue_folder,
                e7g_intermediate,
                conf_params_obj.estimate_agb.intermediate_ground_averaging,
                temp_output_folder,
                gdal_path,
            )

        # LUT: FNF
        lut_fnf_boundaries = np.zeros((len(equi7_fnf_mask_fnames), 4))
        lut_fnf_paths = []
        for fnf_idx, fnf_name_curr in enumerate(equi7_fnf_mask_fnames):

            data_driver = gdal.Open(fnf_name_curr, GA_ReadOnly)
            geotransform_equi7 = data_driver.GetGeoTransform()
            east_len = data_driver.RasterXSize
            north_len = data_driver.RasterYSize

            data_driver = None

            east_min = geotransform_equi7[0]
            east_max = east_min + geotransform_equi7[1] * east_len
            north_in = geotransform_equi7[3]
            north_out = north_in + geotransform_equi7[5] * north_len

            lut_fnf_boundaries[fnf_idx, :] = [
                east_min,
                east_max,
                min(north_in, north_out),
                max(north_in, north_out),
            ]
            lut_fnf_paths.append(fnf_name_curr)

        # To be set from configuration file in the final version
        space_varying_ground_cancellation = False

        ########################## INITIAL STEPS END #############################

        ########################## STACK BASED STEPS ##############################

        # LookUp tables initialization
        number_of_stacks = len(input_params_obj.stack_based_processing.stack_composition.keys())
        lut_progressive_stacks = np.zeros((number_of_stacks, 6))
        lut_stacks_boundaries = np.zeros((number_of_stacks, 4))
        lut_stacks_paths = []

        # sigma0 and theta initialization
        sigma0_equi7_fnames = {}
        theta_equi7_fnames = {}
        # cycle for each stack
        # for stack_key, scene_dict in proc_inputs.stacks_scenes_fnames_orders.items():
        for (stack_progressive_idx, (unique_stack_id, acquisitions_pf_names),) in enumerate(
            input_params_obj.stack_based_processing.stack_composition.items()
        ):

            (
                global_cycle_idx,
                heading_deg,
                rg_swath_idx,
                rg_sub_swath_idx,
                az_swath_idx,
                _,
            ) = decode_unique_acquisition_id_string(unique_stack_id + "_BSL_00")

            # LUT: progresive stacks
            lut_progressive_stacks[stack_progressive_idx, :] = [
                stack_progressive_idx,
                global_cycle_idx,
                heading_deg,
                rg_swath_idx,
                rg_sub_swath_idx,
                az_swath_idx,
            ]

            # make temporary sub-directories
            temp_output_folder_gr = os.path.join(temp_output_folder, "geocoded", unique_stack_id)
            temp_output_folder_e7 = os.path.join(temp_output_folder, "equi7", unique_stack_id)
            os.makedirs(temp_output_folder_gr)
            os.makedirs(temp_output_folder_e7)

            ### load data ( and oversample if requested and if needed )
            try:
                logging.info("AGB: Data loading for stack " + unique_stack_id + "; this may take a while:")

                (beta0_calibrated, master_id, raster_info, raster_info_orig,) = read_and_oversample_data(
                    input_params_obj.dataset_query.L1C_repository,
                    acquisitions_pf_names,
                    conf_params_obj.processing_flags.enable_resampling,
                )

            except Exception as e:
                logging.error("AGB: error during input data reading: " + str(e), exc_info=True)
                raise

            ### load or compute auxiliary data
            try:

                read_ref_h = (
                    not conf_params_obj.processing_flags.apply_calibration_screen
                    and conf_params_obj.processing_flags.DEM_flattening
                )
                read_cal_screens = conf_params_obj.processing_flags.apply_calibration_screen
                geometry_aux_are_present = check_if_geometry_auxiliaries_are_present(
                    input_params_obj.stack_based_processing,
                    unique_stack_id,
                    acquisitions_pf_names,
                    read_ref_h=read_ref_h,
                )

                if conf_params_obj.processing_flags.compute_geometry or not geometry_aux_are_present:

                    # messages for the log:
                    if conf_params_obj.processing_flags.compute_geometry:
                        logging.info("AGB: calling geometry library for stack " + unique_stack_id + "\n")
                        if geometry_aux_are_present:
                            logging.warning("    geometry auxiliaries will be overwritten for stack " + unique_stack_id)
                        else:
                            logging.info("\n")
                    else:
                        logging.warning(
                            'AGB: calling geometry library since AuxiliaryProductsFolder "Geometry" is empty or not complete \n'
                        )

                    (
                        _,
                        _,
                        ellipsoid_slope,
                        _,
                        _,
                        _,
                        sar_geometry_master,
                    ) = compute_and_oversample_geometry_auxiliaries(
                        input_params_obj.dataset_query.L1C_repository,
                        input_params_obj.stack_based_processing,
                        unique_stack_id,
                        acquisitions_pf_names,
                        master_id,
                        conf_params_obj.processing_flags.enable_resampling,
                        force_ellipsoid=True,
                    )

                    (
                        ecef_grid,
                        off_nadir_angle_rad,
                        slope,
                        kz,
                        reference_height,
                        _,
                        _,
                    ) = compute_and_oversample_geometry_auxiliaries(
                        input_params_obj.dataset_query.L1C_repository,
                        input_params_obj.stack_based_processing,
                        unique_stack_id,
                        acquisitions_pf_names,
                        master_id,
                        conf_params_obj.processing_flags.enable_resampling,
                        comp_ref_h=read_ref_h,
                        sar_geometry_master=sar_geometry_master,
                    )

                    logging.info("Geometry library: correcting geometry global / local reference")
                    slope = slope - ellipsoid_slope

                    # The kz taking into account the ground slope is generated too
                    kz_ground_slope = kz.copy()

                    for swath_id in kz.keys():
                        kz[swath_id] = (
                            kz[swath_id]
                            * np.sin(off_nadir_angle_rad[master_id])
                            / np.sin(off_nadir_angle_rad[master_id] - ellipsoid_slope)
                        )
                        kz_ground_slope[swath_id] = kz[swath_id] / (
                            1 + np.tan(slope) / np.tan(off_nadir_angle_rad[master_id] - ellipsoid_slope)
                        )
                    off_nadir_angle_rad[master_id] = off_nadir_angle_rad[master_id] - ellipsoid_slope
                    del ellipsoid_slope

                    logging.info("AGB: ...geometry auxiliaries computation done.")

                else:

                    logging.info(
                        "AGB: geometry auxiliaries are provided from user, so they are now being loaded and not computed, for stack "
                        + unique_stack_id
                        + "\n"
                    )

                    (
                        ecef_grid,
                        off_nadir_angle_rad,
                        slope,
                        kz,
                        reference_height,
                        _,
                        _,
                        _,
                        _,
                        _,
                        _,
                        _,
                    ) = read_and_oversample_aux_data(
                        input_params_obj.stack_based_processing,
                        unique_stack_id,
                        acquisitions_pf_names,
                        conf_params_obj.processing_flags.enable_resampling,
                        raster_info_orig,
                        read_ref_h=read_ref_h,
                    )

                    # The kz taking into account the ground slope is generated too
                    # ellipsoid_slope is set to zero
                    kz_ground_slope = kz.copy()
                    for swath_id in kz.keys():
                        kz_ground_slope[swath_id] = kz[swath_id] / (
                            1 + np.tan(slope) / np.tan(off_nadir_angle_rad[master_id])
                        )
                    logging.info("AGB: ...geometry auxiliaries loading done.")

                # read the rest of auxiliaries which are notpart of the geometry library:
                if read_cal_screens:

                    logging.warning("AGB: loading calibration screens \n")

                    (
                        _,
                        _,
                        _,
                        _,
                        _,
                        _,
                        _,
                        cal_screens,
                        cal_screens_raster_info,
                        _,
                        _,
                        _,
                    ) = read_and_oversample_aux_data(
                        input_params_obj.stack_based_processing,
                        unique_stack_id,
                        acquisitions_pf_names,
                        conf_params_obj.processing_flags.enable_resampling,
                        raster_info_orig,
                        read_cal_screens=read_cal_screens,
                        read_ecef=False,
                        read_off_nadir=False,
                        read_slope=False,
                        read_kz=False,
                        read_ref_h=False,
                        read_dist=False,
                    )
                    logging.info("...done")

            except Exception as e:
                logging.error(
                    "AGB: error during auxiliary data computation and/or loading: " + str(e), exc_info=True,
                )
                raise

            ### Screen calibration (ground steering)
            try:
                if conf_params_obj.processing_flags.apply_calibration_screen:
                    logging.info("AGB: applying calibration screen...")
                    beta0_calibrated = apply_calibration_screens(
                        beta0_calibrated, raster_info, cal_screens, cal_screens_raster_info, master_id,
                    )
                    logging.info("...done.\n")

                elif conf_params_obj.processing_flags.DEM_flattening:
                    logging.info("AGB: DEM flattening... ")
                    beta0_calibrated = apply_dem_flattening(
                        beta0_calibrated, kz, reference_height, master_id, raster_info
                    )
                    logging.info("...done.\n")

            except Exception as e:
                logging.error(
                    "AGB: error during screen calibration or DEM flattening." + str(e), exc_info=True,
                )
                raise

            ### ground notching
            try:

                if space_varying_ground_cancellation:

                    # The map specifying the space-varying height to be emphasized
                    # must be available here. For the moment a map filled with the
                    # value taken from the configuration file is used.
                    current_enhanced_forest_height = slope.copy()
                    current_enhanced_forest_height.fill(conf_params_obj.ground_cancellation.enhanced_forest_height)

                    logging.info("AGB: space-varying ground contribute cancellation...:")
                    DN_beta0_notched = ground_cancellation(
                        beta0_calibrated,
                        kz_ground_slope,
                        conf_params_obj.ground_cancellation.multi_master_flag,
                        current_enhanced_forest_height,
                        conf_params_obj.ground_cancellation.equalization_flag,
                        raster_info.resolution_m_slant_rg,
                        off_nadir_angle_rad[master_id],
                        slope,
                    )  # beta0_calibrated: 3 pol (nrg x Naz  x N images); DN_beta0_notched: 3 pol (Nrg x Naz x 1 image)
                    del beta0_calibrated, kz, kz_ground_slope, current_enhanced_forest_height
                else:

                    logging.info("AGB: ground contribute cancellation...:")

                    DN_beta0_notched = ground_cancellation(
                        beta0_calibrated,
                        kz,
                        conf_params_obj.ground_cancellation.multi_master_flag,
                        conf_params_obj.ground_cancellation.enhanced_forest_height,
                        conf_params_obj.ground_cancellation.equalization_flag,
                        raster_info.resolution_m_slant_rg,
                        off_nadir_angle_rad[master_id],
                        slope,
                    )  # beta0_calibrated: 3 pol (nrg x Naz  x images); DN_beta0_notched: 3 pol (Nrg x Naz x 1 image)

                    del beta0_calibrated, kz

            except Exception as e:
                logging.error("AGB: error during ground cancellation: " + str(e), exc_info=True)
                raise
            logging.info("...done.\n")

            ### compute mean look angle
            look_angle_rad = np.nanmean(off_nadir_angle_rad[master_id])
            logging.info("AGB: look angle used is {} [deg] \n".format(np.rad2deg(look_angle_rad)))

            if (
                conf_params_obj.estimate_agb.intermediate_ground_averaging
                > conf_params_obj.estimate_agb.product_resolution / 2
            ):
                sigma_ground_res_m = conf_params_obj.estimate_agb.product_resolution / 2
                logging.warning(
                    '"intermediate_ground_averaging" cannot be greater than "product_resolution/2", setting it to {}'.format(
                        sigma_ground_res_m
                    )
                )
            else:
                sigma_ground_res_m = conf_params_obj.estimate_agb.intermediate_ground_averaging

            if conf_params_obj.processing_flags.multilook_heading_correction:
                sigma_ground_res_m = resolution_heading_correction(sigma_ground_res_m, heading_deg)

            windtm_x = np.int(np.round(sigma_ground_res_m / raster_info.pixel_spacing_az / 2) * 2 + 1)
            windtm_y = np.int(
                np.round(sigma_ground_res_m / (raster_info.pixel_spacing_slant_rg / np.sin(look_angle_rad)) / 2) * 2 + 1
            )

            sub_factor_x = np.int((windtm_x - 1) / 2)
            sub_factor_y = np.int((windtm_y - 1) / 2)

            logging.info("AGB: multilooking of incidence angle...")
            theta_multi_looked_sr = convolve2d(
                off_nadir_angle_rad[master_id] - slope,
                np.ones((windtm_y, windtm_x)) / windtm_y / windtm_x,
                mode="same",
            )
            theta_multi_looked_sr = theta_multi_looked_sr[::sub_factor_y, ::sub_factor_x]
            logging.info("...done.\n")

            sigma0_sr = {}
            for pol_name in DN_beta0_notched.keys():

                logging.info("AGB: multilooking of ground notched for polarization {}...".format(pol_name))
                beta0_notched_multi_looked = convolve2d(
                    np.absolute(DN_beta0_notched[pol_name]) ** 2,
                    np.ones((windtm_y, windtm_x)) / windtm_y / windtm_x,
                    mode="same",
                )

                logging.info("AGB: sigma0 computation for polarization {}...".format(pol_name))
                sigma0_sr[pol_name] = beta0_notched_multi_looked[::sub_factor_y, ::sub_factor_x] * np.sin(
                    theta_multi_looked_sr
                )

                sigma0_sr[pol_name][sigma0_sr[pol_name] < 0] = np.NaN

                logging.info("...done.\n")

            del beta0_notched_multi_looked

            ### saving breakpoints
            if conf_params_obj.processing_flags.save_breakpoints:
                logging.info("AGB: saving breakpoints (in slant range geometry) on " + breakpoints_output_folder)
                post_string = "_SR_" + unique_stack_id

                breakpoint_names = ["ground_cancelled" + post_string]

                save_breakpoints(breakpoints_output_folder, breakpoint_names, [DN_beta0_notched])
                logging.info("...done.\n")

            az_vec_subs = np.arange(0, raster_info.num_lines, sub_factor_x)
            rg_vec_subs = np.arange(0, raster_info.num_samples, sub_factor_y)

            ### Interpolate it over a regular lat lon grid (with grid data): generate a regular grid for the interpolation by using Max and Min lat lon from the ECEFGRID_LLH (make the vectors a bit longer), for the grid steps use the minimum steps from the ECEFGRID_LLH)
            logging.info(unique_stack_id + ": Geocoding data...")
            try:

                # initialize the geocoding
                min_spacing_m = min(
                    sub_factor_x * raster_info.pixel_spacing_az, sub_factor_y * raster_info.pixel_spacing_slant_rg,
                )
                min_spacing_m = min(min_spacing_m, equi7_sampling_intermediate)

                logging.info("Geocoding spacing set to {} [m]".format(min_spacing_m))

                (
                    lon_in,
                    lat_in,
                    alt_in,
                    lonMeshed_out,
                    latMeshed_out,
                    valid_values_mask,
                    lon_regular_vector,
                    used_lon_step,
                    lat_regular_vector,
                    used_lat_step,
                ) = geocoding_init(ecef_grid, rg_vec_subs, az_vec_subs, min_spacing_m)

                # geocode the sigma0 (three polarizations)
                sigma0_gr = {}
                for pol_name in sigma0_sr.keys():

                    logging.info("AGB: geocoding the sigma0 for polarization {}...".format(pol_name))
                    sigma0_gr[pol_name] = geocoding(
                        sigma0_sr[pol_name], lon_in, lat_in, lonMeshed_out, latMeshed_out, valid_values_mask,
                    )

                    logging.info("...done.\n")

                del sigma0_sr

                # geocode the theta incidence angle
                logging.info("AGB: Geocoding of incidence angle...")
                theta_multi_looked_gr = geocoding(
                    theta_multi_looked_sr, lon_in, lat_in, lonMeshed_out, latMeshed_out, valid_values_mask,
                )
                logging.info("...done.\n")
                del theta_multi_looked_sr

                logging.info("...done.\n")

            except Exception as e:
                logging.error("AGB: error during geocoding: " + str(e), exc_info=True)
                raise

            ### create GEOTIFF of all the layers (estimation, mask ):
            logging.info(unique_stack_id + ": formatting data to GEOTIFF...")
            try:

                upper_left_easting_coord = lon_regular_vector[0]  # i.e. horizontal
                sampling_step_east_west = used_lon_step
                upper_left_northing_coord = lat_regular_vector[0]  # i.e. vertical
                sampling_step_north_south = used_lat_step
                geotransform = [
                    upper_left_easting_coord,
                    sampling_step_east_west,
                    0,
                    upper_left_northing_coord,
                    0,
                    sampling_step_north_south,
                ]

                # geotiff of the sigma0 (three polarizations)
                sigma0_ground_fnames = {}
                for pol_name in sigma0_gr.keys():
                    sigma0_ground_fnames[pol_name] = os.path.join(temp_output_folder_gr, "sigma0_" + pol_name + ".tif",)

                    tiff_formatter(
                        sigma0_gr[pol_name],
                        sigma0_ground_fnames[pol_name],
                        geotransform,
                        gdal_data_format=gdal.GDT_Float32,
                    )

                del sigma0_gr

                # geotiff of the theta
                theta_ground_fname = os.path.join(temp_output_folder_gr, "theta.tif")

                tiff_formatter(
                    theta_multi_looked_gr, theta_ground_fname, geotransform, gdal_data_format=gdal.GDT_Float32,
                )

                del theta_multi_looked_gr
                logging.info("...done.\n")

            except Exception as e:
                logging.error("AGB: error during GEOTIFF formatting: " + str(e), exc_info=True)
                raise

            ### formatting data to EQUI7
            logging.info(unique_stack_id + ": formatting into EQUI7 grid...")
            try:

                sigma0_equi7_fnames[unique_stack_id] = {}
                # equi7 of the sigma0 (three polarizations)
                for pol_name in sigma0_ground_fnames.keys():

                    equi7_sigma0_outdir = os.path.join(temp_output_folder_e7, "sigma0_" + pol_name)

                    logging.info(
                        "image2equi7grid IN: " + sigma0_ground_fnames[pol_name] + " , OUT:" + equi7_sigma0_outdir
                    )
                    sigma0_equi7_fnames[unique_stack_id][pol_name] = image2equi7grid(
                        e7g_intermediate,
                        sigma0_ground_fnames[pol_name],
                        equi7_sigma0_outdir,
                        gdal_path=gdal_path,
                        inband=None,
                        subgrid_ids=None,
                        accurate_boundary=False,
                        withtilenamesuffix=False,
                        resampling_type="bilinear",
                        tile_nodata=np.nan,
                    )

                # equi7 of the theta
                equi7_theta_outdir = os.path.join(temp_output_folder_e7, "theta")

                logging.info("image2equi7grid IN: " + theta_ground_fname + " , OUT:" + equi7_theta_outdir)
                theta_equi7_fnames[unique_stack_id] = image2equi7grid(
                    e7g_intermediate,
                    theta_ground_fname,
                    equi7_theta_outdir,
                    gdal_path=gdal_path,
                    inband=None,
                    subgrid_ids=None,
                    accurate_boundary=False,
                    withtilenamesuffix=False,
                    resampling_type="bilinear",
                    tile_nodata=np.nan,
                )

            except Exception as e:
                logging.error("AGB: error during EQUI7 formatting: " + str(e), exc_info=True)
                raise

            if not os.path.exists(theta_equi7_fnames[unique_stack_id][0]):
                error_message = "EQUI7 grid has not been generated, output is absent "
                logging.error(error_message)
                raise RuntimeError(error_message + " :" + theta_equi7_fnames[unique_stack_id][0])

            # N0: upper left corner for output map (UTM 32S)
            _, x_upper_left, y_upper_left = e7g_intermediate.lonlat2xy(
                geographic_boundaries_per_stack[unique_stack_id].lon_min,
                geographic_boundaries_per_stack[unique_stack_id].lat_max,
            )
            north_in = y_upper_left
            east_min = x_upper_left

            # lower right corner for output map (UTM 32S)
            _, x_lower_right, y_lower_right = e7g_intermediate.lonlat2xy(
                geographic_boundaries_per_stack[unique_stack_id].lon_max,
                geographic_boundaries_per_stack[unique_stack_id].lat_min,
            )
            north_out = y_lower_right
            east_max = x_lower_right

            lut_stacks_boundaries[stack_progressive_idx, :] = [
                east_min,
                east_max,
                min(north_in, north_out),
                max(north_in, north_out),
            ]

            lut_stacks_paths.append(os.path.join(temp_output_folder_e7))

            logging.info("...done.\n")

        equi7_initialization = {
            "equi7_sampling_intermediate": equi7_sampling_intermediate,
            "e7g_intermediate": e7g_intermediate,
        }
        # create the CoreProcessingAGB configuration file, starting from the default one
        # Read the default conf file:
        default_coreprocessing_conf_file = os.path.join(
            os.path.dirname(self.configuration_file), "ConfigurationFile_CoreProcessingAGB_Default.xml",
        )

        # update the conf paths:
        # source composition:
        # source[index_obs][index_stack][index_file][path,band_id]
        for index_obs, name in enumerate(conf_params_obj.estimate_agb.residual_function.formula_observables.name):
            if not conf_params_obj.estimate_agb.residual_function.formula_observables.source_paths[index_obs]:

                if name == "neg_sigma0_hh_db":
                    pol_name = "hh"
                    stack_list = []
                    for index_stack, sigma0_pols_dict in enumerate(sigma0_equi7_fnames.values()):
                        sigma0_file_names_list = sigma0_pols_dict[pol_name]
                        file_list = []
                        for index_file, sigma0_file_name in enumerate(sigma0_file_names_list):
                            layer_list = [sigma0_file_name, 0]
                            file_list.append(layer_list)
                        stack_list.append(file_list)

                    ## temporary (for tests with tomographic data)
                    # layer_list = [r'C:\Users\macie\Documents\bio_input\aux_for_agb_dev\out_tomo_GGS_50m_RES_100m\slice_30_m\EQUI7_AF050M\E045N048T3\cov_vv_30_m_EQUI7_AF050M_E045N048T3.tif', 0]
                    # file_list = [layer_list]
                    # stack_list = [file_list]

                    conf_params_obj.estimate_agb.residual_function.formula_observables.source_paths[
                        index_obs
                    ] = stack_list
                    conf_params_obj.estimate_agb.residual_function.formula_observables.source_resolution[
                        index_obs
                    ] = sigma_ground_res_m
                    conf_params_obj.estimate_agb.residual_function.formula_observables.source_unit[index_obs] = "none"

                elif name == "neg_sigma0_hv_db":
                    pol_name = "vh"
                    stack_list = []
                    for index_stack, sigma0_pols_dict in enumerate(sigma0_equi7_fnames.values()):
                        sigma0_file_names_list = sigma0_pols_dict[pol_name]
                        file_list = []
                        for index_file, sigma0_file_name in enumerate(sigma0_file_names_list):
                            layer_list = [sigma0_file_name, 0]
                            file_list.append(layer_list)
                        stack_list.append(file_list)

                    ## temporary (for tests with tomographic data)
                    # layer_list = [r'C:\Users\macie\Documents\bio_input\aux_for_agb_dev\out_tomo_GGS_50m_RES_100m\slice_30_m\EQUI7_AF050M\E045N048T3\cov_vh_30_m_EQUI7_AF050M_E045N048T3.tif', 0]
                    # file_list = [layer_list]
                    # stack_list = [file_list]

                    conf_params_obj.estimate_agb.residual_function.formula_observables.source_paths[
                        index_obs
                    ] = stack_list
                    conf_params_obj.estimate_agb.residual_function.formula_observables.source_resolution[
                        index_obs
                    ] = sigma_ground_res_m
                    conf_params_obj.estimate_agb.residual_function.formula_observables.source_unit[index_obs] = "none"

                elif name == "neg_sigma0_vv_db":
                    pol_name = "vv"
                    stack_list = []
                    for index_stack, sigma0_pols_dict in enumerate(sigma0_equi7_fnames.values()):
                        sigma0_file_names_list = sigma0_pols_dict[pol_name]
                        file_list = []
                        for index_file, sigma0_file_name in enumerate(sigma0_file_names_list):
                            layer_list = [sigma0_file_name, 0]
                            file_list.append(layer_list)
                        stack_list.append(file_list)

                    ## temporary (for tests with tomographic data)
                    # layer_list = [r'C:\Users\macie\Documents\bio_input\aux_for_agb_dev\out_tomo_GGS_50m_RES_100m\slice_30_m\EQUI7_AF050M\E045N048T3\cov_vv_30_m_EQUI7_AF050M_E045N048T3.tif', 0]
                    # file_list = [layer_list]
                    # stack_list = [file_list]

                    conf_params_obj.estimate_agb.residual_function.formula_observables.source_paths[
                        index_obs
                    ] = stack_list
                    conf_params_obj.estimate_agb.residual_function.formula_observables.source_resolution[
                        index_obs
                    ] = sigma_ground_res_m
                    conf_params_obj.estimate_agb.residual_function.formula_observables.source_unit[index_obs] = "none"

                elif (name == "cos_local_db") or (name == "theta_local"):

                    stack_list = []
                    for index_stack, theta_list in enumerate(theta_equi7_fnames.values()):
                        file_list = []
                        for index_file, theta_file_name in enumerate(theta_list):
                            layer_list = [theta_file_name, 0]
                            file_list.append(layer_list)
                        stack_list.append(file_list)

                    ## temporary (for tests with tomographic data)
                    # layer_list = [r'C:\Users\macie\Documents\bio_input\aux_for_agb_dev\out_tomo_GGS_50m_RES_100m\theta\EQUI7_AF050M\E045N048T3\theta_AF050M_E045N048T3.tif', 0]
                    # file_list = [layer_list]
                    # stack_list = [file_list]

                    conf_params_obj.estimate_agb.residual_function.formula_observables.source_paths[
                        index_obs
                    ] = stack_list
                    conf_params_obj.estimate_agb.residual_function.formula_observables.source_resolution[
                        index_obs
                    ] = sigma_ground_res_m
                    conf_params_obj.estimate_agb.residual_function.formula_observables.source_unit[index_obs] = "rad"

                ## temporary (for tests with gedi-like data for calibration)
                # elif name == "agb_1_cal_1km_db":
                #     # layer_list = [r'C:\Users\macie\Documents\bio_input\aux_for_agb_dev\lope_lidar\lidar_agb\EQUI7_AF050M\E045N048T3\lidar_AGB_1km_AF050M_E045N048T3.tif', 0]
                #     # file_list = [layer_list]
                #     # stack_list = [file_list]
                #     stack_list = []
                #     conf_params_default.AGB.residual_function.formula_observables.source_paths[index_obs] = stack_list
                #     conf_params_default.AGB.residual_function.formula_observables.source_resolution[index_obs] = 1000
                #     conf_params_default.AGB.residual_function.formula_observables.source_unit[index_obs] = 't/ha'

                elif name == "agb_1_cal_db":

                    file_list = []
                    for index_file, cal_file_name in enumerate(lut_cal_paths):
                        layer_list = [cal_file_name, 0]
                        file_list.append(layer_list)
                    stack_list = [file_list]

                    conf_params_obj.estimate_agb.residual_function.formula_observables.source_paths[
                        index_obs
                    ] = stack_list
                    conf_params_obj.estimate_agb.residual_function.formula_observables.source_resolution[
                        index_obs
                    ] = 100
                    conf_params_obj.estimate_agb.residual_function.formula_observables.source_unit[index_obs] = "none"

                    ## uncomment if testing with gedi-like data
                    # stack_list = []
                    conf_params_obj.estimate_agb.residual_function.formula_observables.source_paths[
                        index_obs
                    ] = stack_list
                    conf_params_obj.estimate_agb.residual_function.formula_observables.source_resolution[index_obs] = 50
                    conf_params_obj.estimate_agb.residual_function.formula_observables.source_unit[index_obs] = "t/ha"

                ## temporary (for tests with height data)
                # elif (name == "tomo_h_db"):
                #     layer_list = [r'C:\Users\macie\Documents\bio_input\aux_for_agb_dev\out_tomo_fh\BIOMASS_L2_20201118T123101\TOMO_FH\Products\global_FH\EQUI7_AF050M\E045N048T3\FH_EQUI7_AF050M_E045N048T3.tif', 0]
                #     file_list = [layer_list]
                #     stack_list = [file_list]
                #     conf_params_default.AGB.residual_function.formula_observables.source_paths[index_obs] = stack_list
                #     conf_params_default.AGB.residual_function.formula_observables.source_resolution[index_obs] = 100
                #     conf_params_default.AGB.residual_function.formula_observables.source_unit[index_obs] = 'm'

                elif name == "forest_class":

                    file_list = []
                    for index_file, fnf_file_name in enumerate(lut_fnf_paths):
                        layer_list = [fnf_file_name, 0]
                        file_list.append(layer_list)
                    stack_list = [file_list]

                    conf_params_obj.estimate_agb.residual_function.formula_observables.source_paths[
                        index_obs
                    ] = stack_list
                    conf_params_obj.estimate_agb.residual_function.formula_observables.source_resolution[
                        index_obs
                    ] = 100
                    conf_params_obj.estimate_agb.residual_function.formula_observables.source_unit[index_obs] = "none"

        lut_cal = LookupTableAGB(paths=lut_cal_paths, boundaries=lut_cal_boundaries, progressive=None)
        lut_fnf = LookupTableAGB(paths=lut_fnf_paths, boundaries=lut_fnf_boundaries, progressive=None)
        lut_stacks = LookupTableAGB(
            paths=lut_stacks_paths, boundaries=lut_stacks_boundaries, progressive=lut_progressive_stacks,
        )

        # write the input file with the sections needed by the Core Processing AGB APP:
        out_input_file_xml = os.path.join(
            os.path.dirname(input_params_obj.output_specification.output_folder), "Input_File_CoreProcessingAGB.xml"
        )
        input_params_obj.core_processing_agb = fill_core_processing_agb_obj(
            input_params_obj, lut_cal, lut_fnf, lut_stacks
        )
        write_input_file(input_params_obj, out_input_file_xml)

        # write the updatec conf file:
        coreprocessing_configuration_file_xml = os.path.join(
            os.path.dirname(input_params_obj.output_specification.output_folder), "Configuration_File_paths_updated.xml"
        )
        write_configuration_file(conf_params_obj, coreprocessing_configuration_file_xml)

        end_message = "AGB stack-based processing APP ended correctly.\n"
        logging.info(end_message)
        print(end_message)
        ########################## END OF STACK BASED STEPS ######################

        return out_input_file_xml, coreprocessing_configuration_file_xml


class CoreProcessingAGB(Task):
    """CoreProcessingAGB APP

    CoreProcessingAGB is the second of the two sub-APPs called by AboveGroundBiomass processor.

    It computes the AGB product starting from stack inputs generated by "StackBasedProcessingAGB" APP.
    It is automatically called in the default run (see Getting Started  and Tutorial documentation sections),
    or can be manually launched (stand alone) following sequence:
    dataset_query -> StackBasedProcessingAGB -> CoreProcessingAGB

    Attributes
    ----------
    configuration_file : str
        path of the Configuration_File.xml file
        
    Methods
    -------
    run( input_file_path )
        run the CoreProcessingAGB APP
    name : str
        name of the APP

    See Also
    --------
    biopal.dataset_query.dataset_query.dataset_query : it's the first APP to be called in the manual sequence
    StackBasedProcessingAGB : it's the APP which prepares the input for this APP

    Notes
    -----
    Stand alone "CoreProcessingAGB" APP launch. 
    It is possible to launch many istances of "CoreProcessingAGB" APP without need to re launch all the processor.
    
    The app needs a configuration file updated with object paths (updated during "StackBasedProcessingAGB" APP run)
    and can be customized to run different core processings with same stack based pre processing from StackBasedProcessingAGB.

    Examples
    --------
    Manual AGB chain execution

    >>> from biopal.dataset_query.dataset_query import dataset_query
    >>> from biopal.AGB.main_AGB import StackBasedProcessingAGB, CoreProcessingAGB
    >>> dq_obj = dataset_query()
    >>> input_file_up1 = dq_obj.run( input_file )
    >>> sbp_obj = StackBasedProcessingAGB( config_file )
    >>> input_file_up2, config_file_up = sbp_obj.run( input_file_up1 )

    At this point, the CoreProcessingAGB can be launched with different configuration parameters, modifying config_file_up

    >>> agbp_obj = CoreProcessingAGB( config_file_up )
    >>> agbp_obj.run( input_file_up2 )

    - input_file: path of the BioPAL input file
    - input_file_up1: same of input_file with also the "stack_based_processing" section   
    - input_file_up2: same of input_file_up1 with also the "core_processing_agb" section
    - config_file: path of the BioPAL configuration file 
    - config_file_up: same of config_file, with paths updated
    """

    def __init__(self, configuration_file):

        super().__init__(configuration_file)

    def _run(self, input_file):

        # parse the input file and check for the needed sections in it:
        if isinstance(input_file, str):
            check_if_path_exists(input_file, "FILE")
            input_params_obj = parse_input_file(input_file)
        else:
            input_params_obj = input_file
        if (
            input_params_obj.L2_product is None
            or input_params_obj.output_specification is None
            or input_params_obj.dataset_query is None
            or input_params_obj.core_processing_agb is None
        ):
            error_message = [
                "CoreProcessingAGB APP input file should contain at least"
                ' "L2_product", "output_specification", "dataset_query" and "core_processing_agb" sections'
            ]
            raise InvalidInputError(error_message)

        # parse the configuration file and check for the needed sections in it:
        if isinstance(input_file, str):
            check_if_path_exists(self.configuration_file, "FILE")
            conf_params_obj = parse_configuration_file(self.configuration_file)
        else:
            conf_params_obj = self.configuration_file
        if (
            conf_params_obj.processing_flags is None
            or conf_params_obj.ground_cancellation is None
            or conf_params_obj.estimate_agb is None
        ):
            error_message = [
                "Configuration file for  CoreProcessingAGB APP should contain at least"
                ' "processing_flags", "ground_cancellation" and "estimate_agb" sections'
            ]
            raise InvalidInputError(error_message)

        # Initialize the logger, of the APP has been called in a stand-alone call
        if not hasattr(logging.getLoggerClass().root.handlers[0], "baseFilename"):
            start_logging(
                os.path.dirname(input_params_obj.output_specification.output_folder),
                input_params_obj.L2_product,
                "DEBUG",
                app_name="CoreProcessingAGB",
            )
        logging.info("AGB core-processing APP starting\n")

        # get needed parameters from input and configuration files
        gdal_path, _ = set_gdal_paths(conf_params_obj.gdal.gdal_path, conf_params_obj.gdal.gdal_environment_path)

        geographic_boundaries = input_params_obj.stack_based_processing.geographic_boundaries
        lut_cal_paths = input_params_obj.core_processing_agb.lut_cal.paths
        lut_cal = input_params_obj.core_processing_agb.lut_cal.boundaries

        lut_fnf_paths = input_params_obj.core_processing_agb.lut_fnf.paths
        lut_fnf = input_params_obj.core_processing_agb.lut_fnf.boundaries

        lut_stacks_paths = input_params_obj.core_processing_agb.lut_stacks.paths
        lut_stacks_boundaries = input_params_obj.core_processing_agb.lut_stacks.boundaries
        lut_progressive_stacks = input_params_obj.core_processing_agb.lut_stacks.progressive

        ### initialize the equi7 sampling grid
        equi7_sampling_intermediate = choose_equi7_sampling(
            conf_params_obj.estimate_agb.intermediate_ground_averaging,
            conf_params_obj.estimate_agb.intermediate_ground_averaging / 2,
        )
        e7g_intermediate = Equi7Grid(equi7_sampling_intermediate)

        # setting up directories and making sure that preprocessing has been run
        products_folder = os.path.join(input_params_obj.output_specification.output_folder, "Products")
        temp_proc_folder = os.path.join(products_folder, "temp")
        if not (os.path.exists(temp_proc_folder)):
            error_message = '"temp" folder is not present in output: StackBasedProcessingAGB APP should be launched before CoreProcessingAGB '
            logging.error(error_message)
            raise RuntimeError(error_message)

        # preparing folders for final and temporary output
        global_agb_folder = os.path.join(products_folder, "global_AGB")
        temp_agb_folder = os.path.join(temp_proc_folder, "agb_estimation")
        if not os.path.exists(global_agb_folder):
            os.makedirs(global_agb_folder)
        if not os.path.exists(temp_agb_folder):
            os.makedirs(temp_agb_folder)

        ### THIS PART READS THE SETUP

        """
        
        Note: the xml file with setup and the associated reading function must
        be updated so that the formula, observables, and parameters all are defined
        
        
        """
        # read and initialize all the parameters needed for the inversion
        (
            _,
            first_pixel_north,
            first_pixel_east,
            last_pixel_north,
            last_pixel_east,
            pixel_size_east,
            pixel_size_north,
            sample_spacing_east,
            sample_spacing_north,
            sample_size_east,
            sample_size_north,
            block_spacing_east,
            block_spacing_north,
            block_size_east,
            block_size_north,
            number_of_subsets,
            geographic_grid_sampling,
            sub_grid_string,
        ) = initialize_inversion_parameters(
            equi7_sampling_intermediate,
            input_params_obj.output_specification.geographic_grid_sampling,
            geographic_boundaries,
            conf_params_obj.estimate_agb,
        )

        #### read formula, parameter and observable defitions
        algorithm_setup = conf_params_obj.estimate_agb
        formula_terms = algorithm_setup.residual_function.formula_terms
        formula_parameters = algorithm_setup.residual_function.formula_parameters
        formula_observables = algorithm_setup.residual_function.formula_observables

        # list with paths for images that will be merged at the end
        parameter_map_pathlists = [[] for formula_parameter_name in formula_parameters.name]

        # convert parameter limits from output units to input units
        # (output limits are easier for the user to define as they are more natural;
        # the code will use transformed limits though)
        for idx, transform in enumerate(formula_parameters.transform):
            if transform != "none":
                formula_parameters.limits[idx][0] = transform_function(
                    formula_parameters.limits[idx][0], [-np.inf, np.inf], transform, do_forward=True,
                )
                formula_parameters.limits[idx][1] = transform_function(
                    formula_parameters.limits[idx][1], [-np.inf, np.inf], transform, do_forward=True,
                )
                formula_parameters.limit_units[idx] = transform + "_" + formula_parameters.limit_units[idx]

        # convert  observable limits to radians if in degrees
        for idx, unit in enumerate(formula_observables.limit_units):
            # if there is just one unit, check if the source unit is rad and limits unit is deg; if so, convet the limits unit
            if unit.lower() == "deg":
                formula_observables.limits[idx][0] = np.deg2rad(formula_observables.limits[idx][0])
                formula_observables.limits[idx][1] = np.deg2rad(formula_observables.limits[idx][1])
                formula_observables.limit_units[idx] = "rad"

        # check source units agains limit units, flag discrepancies:
        for unit_source, unit_limit, name in zip(
            formula_observables.source_unit, formula_observables.limit_units, formula_observables.name,
        ):
            if unit_source.lower() != unit_limit.lower():
                logging.warning(
                    "AGB: source unit and limit unit do not match for parameter {} ({} and {}). Could still be OK. Proceed with caution.".format(
                        name, unit_source, unit_limit
                    )
                )

        # check parameter limit units against associated observable limit units, flag discrepancies
        for unit_limit, associated_observable, name in zip(
            formula_parameters.limit_units, formula_parameters.associated_observable_name, formula_parameters.name,
        ):
            if associated_observable != "none":
                observable_idx = np.where(np.array(formula_observables.name) == associated_observable)[0][0]
                if formula_observables.limit_units[observable_idx].lower() != unit_limit.lower():
                    logging.warning(
                        "AGB: limit units for parameter {} and associated observable {} do not match ({} and {}). However, this could still be OK: the observable source unit is {} and transform is {}. Proceed with caution.".format(
                            name,
                            associated_observable,
                            unit_limit,
                            formula_observables.limit_units[observable_idx],
                            formula_observables.source_unit[observable_idx],
                            formula_observables.transform[observable_idx],
                        )
                    )

        # checking that none of observable and parameter names can be contained within each other
        # (parsing of the formula will fail if this is the case)
        defined_quantity_names = formula_parameters.name + formula_observables.name
        for defined_quantity_name_1 in defined_quantity_names:
            for defined_quantity_name_2 in defined_quantity_names:
                if (defined_quantity_name_1 != defined_quantity_name_2) & (
                    (defined_quantity_name_1.find(defined_quantity_name_2) >= 0)
                    | (defined_quantity_name_2.find(defined_quantity_name_1) >= 0)
                ):
                    logging.error(
                        "AGB: parameter and observable names must be unique and cannot contain each other ({} and {} do not fulfill this requirement).".format(
                            defined_quantity_name_1, defined_quantity_name_2
                        )
                    )

        ### PREPARE EQUI7 GRID
        # output equi7 projection info
        #   here, we assume that the output tile and subtile will be that of the first observable source
        #   that is covered by the current block
        equi7_info_source_path = [
            x[0] for z in formula_observables.source_paths for y in z for x in y if x[0].find("EQUI7") >= 0
        ][0]
        equi7_subtile_name = equi7_info_source_path.split(os.path.sep)[-3:-1][0][6:]
        equi7_tile_name = equi7_info_source_path.split(os.path.sep)[-3:-1][1]

        equi7_subgrid_code = equi7_subtile_name[:2]
        equi7_projection_string = get_projection_from_path(equi7_info_source_path)
        equi7_product = Equi7Grid(geographic_grid_sampling)

        ### PREPARE ADDITIONAL SAMPLING POLYGONS
        # (e.g., GEDI pixels or 3rd party reference data)
        additional_sampling_polygons = []
        # this is temporary; this should be read from an xml-file or something
        test_additional_polygons = False
        if test_additional_polygons:
            # as an example, use 1km resolution cells (e.g., to calibrate with GEDI pixels)
            dd = 1000  # dimension
            east_mesh_flattened, north_mesh_flattened = [
                x.flatten()
                for x in np.meshgrid(
                    np.arange(first_pixel_east, last_pixel_east + 1000, 1000),
                    np.arange(last_pixel_north, first_pixel_north + 1000, 1000),
                )
            ]
            for e, n in zip(east_mesh_flattened, north_mesh_flattened):
                additional_sampling_polygons.append(Polygon([(e, n), (e, n + dd), (e + dd, n + dd), (e + dd, n)]))

        ### PREPARING STACK INFO
        # read acquisition info table
        stack_info_table = lut_progressive_stacks
        stack_info_table_columns = [
            "stack_id",
            "global_cycle_id",
            "heading_degrees",
            "swath_id",
            "subswath_id",
            "azimuth_id",
        ]

        ### PREPARING PARAMETER GRID
        # parameter block mesh and flattened coordinate vectors
        block_mesh_east, block_mesh_north = np.meshgrid(
            np.arange(first_pixel_east, last_pixel_east, block_spacing_east),
            np.arange(first_pixel_north, last_pixel_north, block_spacing_north),
        )
        block_corner_coordinates_east, block_corner_coordinates_north = [
            np.float64(x.flatten()) for x in [block_mesh_east, block_mesh_north]
        ]
        # count the blocks and set up a "finished" flag vector
        number_of_blocks = len(block_corner_coordinates_east.flatten())

        # Compute the block processing order
        block_order = compute_block_processing_order(
            block_corner_coordinates_east,
            block_corner_coordinates_north,
            block_size_east,
            block_size_north,
            lut_cal,  # right now, this uses boundaries but in the future it should be capable of using polygons (including additional_sampling_polygons)
            lut_stacks_boundaries,  # right now, this uses boundaries but in the future it should be capable of using polygons
        )

        ### RUNNING PARAMETER BLOCKS
        # block status vector
        block_status = np.zeros(number_of_blocks)
        # flag explanation
        # -1: critical error
        # 0: block run but skipped for some detected, data-related reason
        # 1: block successfully finished

        for counter_blocks_run, current_block_index in enumerate(block_order):

            # flag indicating whether to skip the current block or not
            # (must be reset upon each iteration)
            skip_current_block = False

            logging.info(
                "AGB: Running block {} out of {} (block ID: {})".format(
                    counter_blocks_run + 1, number_of_blocks, current_block_index
                )
            )

            # %% ### CREATING SAMPLING GRID AND TESTING FOR DATA
            try:

                logging.info("AGB: preparing sampling grid and tabulating data...")

                # extent of current block # [min_east, max_east, min_north, max_north]
                current_block_extents = np.array(
                    [
                        block_corner_coordinates_east[current_block_index],
                        block_corner_coordinates_east[current_block_index] + block_size_east,
                        block_corner_coordinates_north[current_block_index],
                        block_corner_coordinates_north[current_block_index] + block_size_north,
                    ]
                )

                sampling_axis_east = np.arange(current_block_extents[0], current_block_extents[1], sample_spacing_east,)
                sampling_axis_north = np.arange(
                    current_block_extents[2], current_block_extents[3], sample_spacing_north,
                )
                east_mesh_flattened, north_mesh_flattened = [
                    x.flatten() for x in np.meshgrid(sampling_axis_east, sampling_axis_north)
                ]

                # prepare main sampling polygons
                main_sampling_polygons = []
                for e, n in zip(east_mesh_flattened.flatten(), north_mesh_flattened.flatten()):
                    main_sampling_polygons.append(
                        Polygon(
                            [
                                (e, n),
                                (e, n + sample_size_east),
                                (e + sample_size_east, n + sample_size_north),
                                (e + sample_size_north, n),
                            ]
                        )
                    )

                # pixel axes for current block
                pixel_axis_east = np.arange(current_block_extents[0], current_block_extents[1], pixel_size_east)
                pixel_axis_north = np.arange(current_block_extents[2], current_block_extents[3], pixel_size_north)

                # geo transform info
                current_geotransform = [
                    pixel_axis_east[0],
                    pixel_size_east,
                    0,
                    pixel_axis_north[0],
                    0,
                    pixel_size_north,
                ]

                # this should be updated to a more generalised approach based on polygons defining the extent of each image
                (block_has_data, _) = check_block_for_data_and_cal(
                    current_block_extents, lut_stacks_boundaries, lut_cal
                )

                # tabulation
                (
                    observable_table,
                    observable_names,
                    identifier_table,
                    identifier_names,
                    parameter_position_table,
                    parameter_position_table_columns,
                    parameter_tables,
                    parameter_table_columns,
                    sample_info_table,
                    sample_info_table_columns,
                ) = sample_and_tabulate_data(
                    current_block_extents,
                    pixel_axis_east,
                    pixel_axis_north,
                    main_sampling_polygons + additional_sampling_polygons,
                    block_has_data,
                    stack_info_table,
                    stack_info_table_columns,
                    formula_observables,
                    algorithm_setup.forest_class_observable_name,
                    formula_parameters,
                    number_of_subsets,
                )

                # checking if the tables contain any data
                if observable_table.shape[0] == 0:
                    logging.info("... skipping block #{} due to lack of valid data points".format(current_block_index))
                    skip_current_block = True
                    block_status[counter_blocks_run] = 0

            except Exception as e:
                logging.error(
                    "AGB: error during data sampling and tabulation." + str(e), exc_info=True,
                )
                block_status[counter_blocks_run] = -1
                raise

            if skip_current_block:
                continue

            # %% ### FITTING MODEL TO TABULATED DATA  AND SAVNG

            try:

                logging.info("AGB: fitting model to data...")

                (
                    parameter_tables,
                    space_invariant_parameter_table,
                    space_invariant_parameter_names,
                    space_variant_parameter_table,
                    space_variant_parameter_names,
                ) = fit_formula_to_random_subsets(
                    formula_terms,
                    number_of_subsets,
                    observable_table,
                    observable_names,
                    identifier_table,
                    identifier_names,
                    parameter_position_table,
                    formula_parameters.name,
                    parameter_tables,
                    parameter_table_columns,
                    formula_parameters.parameter_variabilities,
                    algorithm_setup.fraction_of_cal_per_test,
                    algorithm_setup.fraction_of_roi_per_test,
                    algorithm_setup.min_number_of_cals_per_test,
                    algorithm_setup.min_number_of_rois_per_test,
                    algorithm_setup.transfer_function_name,
                )

                # checking if the tables contain any data
                if len(space_invariant_parameter_table) == 0:
                    logging.info("... skipping block #{} due to no subsets found.".format(current_block_index))
                    skip_current_block = True
                    block_status[counter_blocks_run] = 0
                else:
                    logging.info("AGB: saving data to tables.")
                    # save data
                    parameter_property_names = parameter_table_columns[0][-number_of_subsets - 4 :]
                    line_number_string = ["row"]

                    # select formatting for the output tables
                    table_delimiter = "\t"
                    table_precision = 3
                    table_column_width = 25
                    data_type_lut = [
                        [
                            line_number_string,
                            sample_info_table_columns,
                            identifier_names,
                            parameter_position_table_columns,
                            parameter_property_names,
                            observable_names,
                            formula_parameters.name,
                        ],
                        ["d", "d", "d", "d", "f", "f", "f"],
                    ]

                    # save results table (with parameter estimates replicated across samples etc, if necessary; no indices connecting these to parameter tables)
                    curr_path = os.path.join(temp_agb_folder, "results_table_block_{}.txt".format(current_block_index),)
                    curr_table = np.column_stack(
                        (
                            np.arange(observable_table.shape[0]),
                            sample_info_table,
                            identifier_table,
                            observable_table,
                            space_invariant_parameter_table,
                            space_variant_parameter_table,
                            parameter_position_table,
                        )
                    )
                    curr_column_names = (
                        line_number_string
                        + sample_info_table_columns
                        + identifier_names
                        + observable_names
                        + space_invariant_parameter_names
                        + space_variant_parameter_names
                        + parameter_position_table_columns
                    )
                    save_human_readable_table(
                        curr_path,
                        curr_table,
                        curr_column_names,
                        data_type_lut,
                        table_delimiter,
                        table_precision,
                        table_column_width,
                    )
                    curr_dir = os.path.join(temp_agb_folder, "parameter_estimates_subsets")
                    if not os.path.exists(curr_dir):
                        os.makedirs(curr_dir)
                    # save parameter tables (minimalistic tables with only the most necessary info)
                    for parameter_idx, parameter_name in enumerate(formula_parameters.name):
                        curr_path = os.path.join(
                            curr_dir, "parameter_{}_table_block_{}.txt".format(parameter_name, current_block_index),
                        )
                        curr_table = np.column_stack(
                            (np.arange(parameter_tables[parameter_idx].shape[0]), parameter_tables[parameter_idx],)
                        )
                        curr_column_names = np.concatenate(
                            (np.array(line_number_string), parameter_table_columns[parameter_idx],)
                        )
                        save_human_readable_table(
                            curr_path,
                            curr_table,
                            curr_column_names,
                            data_type_lut,
                            table_delimiter,
                            table_precision,
                            table_column_width,
                        )

            except Exception as e:
                logging.error(
                    "AGB: error during parameter estimation or data saving." + str(e), exc_info=True,
                )
                block_status[counter_blocks_run] = -1
                raise

            if skip_current_block:
                continue

            # %% ### READING IMAGE DATA
            try:

                logging.info("AGB: reading data images.")

                # this needs to be updated in line with the tabulation function
                (
                    forest_class_3d,
                    observables_3d,
                    observables_3d_names,
                    space_invariant_parameters_3d,
                    space_invariant_parameters_3d_names,
                    identifiers_3d,
                    identifiers_3d_names,
                ) = read_and_organise_3d_data(
                    current_block_extents,
                    block_has_data,
                    pixel_axis_north,
                    pixel_axis_east,
                    stack_info_table,
                    stack_info_table_columns,
                    observable_names,
                    formula_observables.source_paths,
                    formula_observables.transform,
                    formula_observables.averaging_method,
                    formula_observables.limits,
                    formula_observables.is_required,
                    [[x, 0] for x in lut_fnf_paths],
                    lut_fnf,
                    identifier_table[:, 2],
                    identifier_table[:, 1],
                    space_invariant_parameter_table,
                    space_invariant_parameter_names,
                    mask_out_area_outside_block=True,
                )

                # skipping to next if all parameters and/or forest class data are missing
                if np.all(forest_class_3d == 0) | np.all(
                    np.array(
                        [
                            np.all(np.isnan(curr_space_invariant_parameter_3d))
                            for curr_space_invariant_parameter_3d in space_invariant_parameters_3d
                        ]
                    )
                ):
                    logging.info("... skipping block #{} due to no image data read.".format(current_block_index))
                    skip_current_block = True
                    block_status[counter_blocks_run] = 0

            except Exception as e:
                logging.error("AGB: error during image reading." + str(e), exc_info=True)
                block_status[counter_blocks_run] = -1
                raise

            if skip_current_block:
                continue

            # %% ### ESTIMATING SPACE-VARIANT PARAMTERS

            try:

                logging.info("AGB: creating space variant parameter images.")

                # detect observables without data (used to remove formula terms that would generate nans)
                observables_with_partial_data = [
                    observable_name
                    for is_nan, observable_name in zip(np.any(np.isnan(observable_table), axis=0), observable_names)
                    if is_nan
                ]
                terms_with_nan_observables = np.any(
                    match_string_lists(formula_terms.string, observables_with_partial_data) >= 0, axis=1,
                )

                if np.any(terms_with_nan_observables):
                    logging.warning(
                        "AGB: skipping formula terms: {} due to lack of useful data for observables: {}.".format(
                            ", ".join(["%d" % (ii + 1) for ii in np.where(terms_with_nan_observables)[0]]),
                            ", ".join(observables_with_partial_data),
                        )
                    )

                #### select relevant formula and weights for this step
                terms_with_zero_weight = np.array(formula_terms.formula_weights.step2) == 0
                if np.any(terms_with_zero_weight):
                    logging.warning(
                        "AGB: skipping formula terms: {} due to zero weights.".format(
                            ", ".join(["%d" % (ii + 1) for ii in np.where(terms_with_zero_weight)[0]])
                        )
                    )

                terms_to_take = ~(terms_with_nan_observables | terms_with_zero_weight)
                formula = subset_iterable(formula_terms.string, terms_to_take)
                formula_weights = subset_iterable(formula_terms.formula_weights.step2, terms_to_take)

                # observables_with_partial_data = [observable_name for is_nan,observable_name in zip(np.any(np.isnan(observable_table),axis=0),observable_names) if is_nan]

                # take out the observables that are in formula and not among space variant parameters
                parameters_for_mapping = np.any(
                    match_string_lists(formula, formula_parameters.name) >= 0, axis=0
                ) & ~np.any(match_string_lists(space_invariant_parameter_names, formula_parameters.name) >= 0, axis=0,)

                observables_for_mapping = np.all(
                    match_string_lists(observable_names, observables_with_partial_data) == -1, axis=1,
                )

                if np.sum(parameters_for_mapping) > 1:
                    logging.error("AGB: the current implementation requires only one space-variant parameter.")

                (space_variant_parameters_3d, space_variant_parameters_3d_names,) = map_space_variant_parameters(
                    formula,
                    formula_weights,
                    forest_class_3d,
                    subset_iterable(observables_3d, observables_for_mapping),
                    subset_iterable(observables_3d_names, observables_for_mapping),
                    space_invariant_parameters_3d,
                    space_invariant_parameters_3d_names,
                    identifiers_3d,
                    identifiers_3d_names,
                    subset_iterable(formula_parameters.name, parameters_for_mapping, False),
                    subset_iterable(formula_parameters.parameter_variabilities, parameters_for_mapping, False,),
                    subset_iterable(formula_parameters.limits, parameters_for_mapping, False),
                    algorithm_setup.transfer_function_name,
                )

                # skipping to next if all parameters and/or forest class data are missing
                if np.all(
                    np.array(
                        [
                            np.all(np.isnan(curr_space_variant_parameter_3d))
                            for curr_space_variant_parameter_3d in space_variant_parameters_3d
                        ]
                    )
                ):
                    logging.info(
                        "... skipping block #{} due to no space variant parameter maps created".format(
                            current_block_index
                        )
                    )
                    skip_current_block = True
                    block_status[counter_blocks_run] = 0

            except Exception as e:
                logging.error(
                    "AGB: error during space variant parameter mapping." + str(e), exc_info=True,
                )
                block_status[counter_blocks_run] = -1
                raise

            if skip_current_block:
                continue

            # %% ### ESTIMATING OTHER PARAMETERS

            try:

                # this is where other things such as error parameters are estimated
                logging.info("AGB: estimating other parameters to be implemented... (includes estimation of error)")

                # for now, the three error elements are hard coded
                additional_parameters_3d = []
                additional_parameters_3d_names = []

            except Exception as e:
                logging.error(
                    "AGB: error during estimation of other parameters." + str(e), exc_info=True,
                )
                block_status[counter_blocks_run] = -1
                raise

            if skip_current_block:
                continue

            # %% ### SAVING SPACE-VARIANT PARAMTERS:

            try:
                logging.info("AGB: saving selected parameters to images.")

                # combine all available mapped parameters
                all_parameters_3d_names = (
                    space_invariant_parameters_3d_names
                    + space_variant_parameters_3d_names
                    + additional_parameters_3d_names
                )
                all_parameters_3d = (
                    space_invariant_parameters_3d + space_variant_parameters_3d + additional_parameters_3d
                )

                for parameter_idx, parameter_name in enumerate(formula_parameters.name):
                    if formula_parameters.save_as_map[parameter_idx]:

                        current_image_to_write = all_parameters_3d[
                            np.where(np.array(all_parameters_3d_names) == parameter_name)[0][0]
                        ]

                        current_file_path = os.path.join(temp_agb_folder, parameter_name)

                        if formula_parameters.transform[parameter_idx] != "none":
                            current_image_to_write = transform_function(
                                current_image_to_write,
                                formula_parameters.limits[parameter_idx],
                                formula_parameters.transform[parameter_idx],
                                do_forward=False,
                            )
                            current_file_path += "_backtransf"

                        current_file_path += (
                            "_block_%d" % (current_block_index) + ".tif"
                        )  # + '_'+equi7_subtile_name+'_'+equi7_tile_name

                        # save tiff file
                        current_file_path = tiff_formatter(
                            [x for x in current_image_to_write.transpose([2, 0, 1])],
                            current_file_path,
                            current_geotransform,
                            gdal_data_format=gdal.GDT_Float32,
                            projection=equi7_projection_string,
                            multi_layers_tiff=True,
                        )

                        # get extents of the current EQUI7 tile
                        # [(left, lower), (right, upper)]
                        try:
                            lon_min, lat_min = getattr(e7g_intermediate, equi7_subgrid_code).xy2lonlat(
                                min(pixel_axis_east), min(pixel_axis_north)
                            )
                            lon_max, lat_max = getattr(e7g_intermediate, equi7_subgrid_code).xy2lonlat(
                                max(pixel_axis_east), max(pixel_axis_north)
                            )
                        except Exception as e:
                            logging.error(
                                'Cannot recognize input FNF Equi7 mask "{}" sub-grid folder name :'.format(
                                    equi7_subgrid_code
                                )
                                + str(e),
                                exc_info=True,
                            )
                            raise

                        output_equi7_file_path = image2equi7grid(
                            equi7_product,
                            current_file_path,
                            temp_agb_folder,
                            gdal_path=gdal_path,
                            ftiles=equi7_product.search_tiles_in_roi(bbox=[(lon_min, lat_min), (lon_max, lat_max)]),
                            accurate_boundary=False,
                            withtilenamesuffix=False,
                            tile_nodata=np.nan,
                        )

                        if formula_parameters.associated_observable_name[parameter_idx] != "none":
                            current_position_in_observable_vector = np.where(
                                match_string_lists(
                                    observable_names, [formula_parameters.associated_observable_name[parameter_idx]],
                                ).flatten()
                                >= 0
                            )[0][0]
                            # current_resolution = 50 # placeholder
                            # current_unit = formula_parameters.units[parameter_idx]
                            if len(formula_observables.source_paths[current_position_in_observable_vector]) == 0:
                                formula_observables.source_paths[current_position_in_observable_vector].append(
                                    [[output_equi7_file_path[0], 0]]
                                )
                            elif len(formula_observables.source_paths[current_position_in_observable_vector]) == 1:
                                formula_observables.source_paths[current_position_in_observable_vector][0].append(
                                    [output_equi7_file_path[0], 0]
                                )
                        parameter_map_pathlists[parameter_idx].append([output_equi7_file_path[0], 0])

                        # lut_cal_paths.append(output_file_path)
                lut_cal = np.row_stack(
                    (lut_cal, np.concatenate((current_block_extents[np.array([0, 1, 3, 2])], np.zeros(1))),)
                )

                skip_current_block = False
                block_status[counter_blocks_run] = 1

            except Exception as e:
                logging.error("AGB: error during saving of maps." + str(e), exc_info=True)
                raise

        # %% FINAL MERGING OF THE IMAGES
        try:
            logging.info("AGB: creating wall-to-wall maps...")
            for parameter_idx, parameter_name in enumerate(formula_parameters.name):
                if formula_parameters.save_as_map[parameter_idx]:

                    tiles_to_save = {}
                    for current_source in parameter_map_pathlists[parameter_idx]:
                        subtile_name = current_source[0].split(os.path.sep)[-3:-1][0][6:]
                        tile_name = current_source[0].split(os.path.sep)[-3:-1][1]

                        if tile_name in tiles_to_save.keys():
                            tiles_to_save[tile_name].append(current_source)
                        else:
                            tiles_to_save[tile_name] = [current_source]

                    for tile_name, tile_sources in tiles_to_save.items():

                        driver = gdal.Open(tile_sources[0][0], GA_ReadOnly)
                        data_merged = np.nan * np.zeros((driver.RasterYSize, driver.RasterXSize))
                        driver = None

                        for idx_tile, current_tile_source in enumerate(tile_sources):

                            # merging current tile blocks togeter:
                            driver = gdal.Open(current_tile_source[0], GA_ReadOnly)

                            if idx_tile == 0:
                                geotransform_out = driver.GetGeoTransform()
                            elif geotransform_out != driver.GetGeoTransform():
                                err_str = "Same equi7 tiles cannot have different geotrasform"
                                logging.error(err_str)
                                raise ValueError(err_str)

                            data_merged = np.nanmean(
                                np.dstack(
                                    (data_merged, driver.GetRasterBand(current_tile_source[1] + 1).ReadAsArray(),)
                                ),
                                axis=2,
                            )

                            driver = None

                        current_merged_file_path = os.path.join(
                            global_agb_folder, equi7_subtile_name, tile_name, parameter_name,
                        )

                        if formula_parameters.transform[parameter_idx] != "none":
                            # current_position_in_observable_vector = parameter_position_in_observable_vector[parameter_idx]
                            # if current_position_in_observable_vector != -1:
                            current_merged_file_path += "_backtransf_"
                        current_merged_file_path += ".tif"

                        if not os.path.exists(os.path.dirname(current_merged_file_path)):
                            os.makedirs(os.path.dirname(current_merged_file_path))

                        current_merged_file_path = tiff_formatter(
                            [data_merged],
                            current_merged_file_path,
                            geotransform_out,
                            gdal_data_format=gdal.GDT_Float32,
                            projection=equi7_projection_string,
                            multi_layers_tiff=True,
                        )
                        logging.info(
                            "... successful for parameter '{}' and filename '{}'".format(
                                parameter_name, current_merged_file_path
                            )
                        )

            end_message = "AGB core-processing APP ended correctly.\n"
            logging.info(end_message)
            print(end_message)

        except Exception as e:
            logging.error(
                "AGB: core-processing APP error during creation of wall-to-wall maps." + str(e), exc_info=True,
            )
            raise


def fill_core_processing_agb_obj(input_params_obj, lut_cal, lut_fnf, lut_stacks):

    """
    Internal function called by the StackBasedProcessingAGB APP:
        
        the StackBasedProcessingAGB APP fills the structure
        to be written into the xml input file for the next APP, which is the 
        core processing for the AGB.
        The returned object "core_processing_agb_obj" contains three lookup tables with paths, boundaries 
        and other specific parameters, of the products (ground notched stacks, 
        forest mask, calibrations) computed by the APP itself.
        
        Usage of the returned object:
            The returned object can be added to the input_params_obj and it 
            can be written to disk if needed, as (core_processing_agb_obj is 
                                                  overwritten by this command, if already present):
                - input_params_obj.core_processing_agb_obj = core_processing_agb_obj
                - write_input_file(input_params_obj, input_file_xml)
    """

    core_processing_agb_obj = core_processing_agb(lut_cal, lut_fnf, lut_stacks,)

    return core_processing_agb_obj



# SPDX-FileCopyrightText: BioPAL <biopal@esa.int>
# SPDX-License-Identifier: MIT

from scipy.constants import c as LIGHTSPEED

# maimum value of the earth rasius (at equator)
MAX_EARTH_RADIUS_METERS = 6378000

# Earth centred, earth fixed, righthanded 3D coordinate system,
# consisting of 3 orthogonal axes with X and Y axes in the equatorial plane,
# positive Z-axis parallel to mean earth rotation axis and pointing towards
# North Pole. UoM: m.
EPSG_CODE_ECEF = "EPSG:4978"

# Ellipsoidal 2D CS. Axes: latitude, longitude. Orientations: north, east. UoM: degree
EPSG_CODE_LLA = "EPSG:4326"

# Oversampling factor to be aplied to input data and auxiliary data when
# resolution / pixel_spacing is < 2
OVERSAMPLING_FACTOR = 2



# -*- coding: utf-8 -*-
import sys
from pathlib import Path
from biopal.__init__ import __version__

sys.path.insert(0, str(Path(__file__).parent.parent.absolute()))

# Build instructions
#   Build latex pdf:
#       python -m sphinx -b latex -a doc doc/_buildlatex
#       cd doc/_buildlatex
#       make.bat
#   Build API Doc html:
#       nox -r -fb conda -s build_doc

# -- Project information -----------------------------------------------------
project = "BioPAL"  # This will be the title of sphinx latex pdf: if empty the pdf title will be the title of index.rst (see html_title for API doc title)
author = "BioPAL team"
copyright = "2021, BioPAL team"

# -- General configuration ---------------------------------------------------
extensions = [
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
	"sphinx.ext.graphviz",
    "numpydoc",
    "nbsphinx",
	
]
autosummary_generate = True
exclude_patterns = ["_build", "**.ipynb_checkpoints", "legacy"]

# -- HTML options ------------------------------------------------------------
html_title = "BioPAL"  # This will be the title for the api doc html
html_logo = "_static/logo.png"
html_static_path = ["_static"]
html_theme = "furo"

# -- LATEX options ------------------------------------------------------------
latex_engine = "xelatex"
latex_logo = "_static/logo.png"

# make title is for the cover page:
latex_maketitle = r"""
\begin{titlepage}

\raggedleft % those are for align all that follows on the right
\sphinxlogo
Generic string on the right \par % "par is used for carriage return"

\centering  % those are for align all that follows on the center
\sphinxlogo
Generic string centred \par

\raggedright   % those are for align all that follows on the left
\sphinxlogo
Generic string on the left \par
\tiny Generic string on the left tiny\par
\small Generic string on the left small\par
\large Generic string on the left large\par
\huge Generic string on the left huge\par
\vspace{5mm} % this generates 5mm of vertical space
\today \par
\vspace{5mm}
BioPAL Team \par

\end{titlepage}
"""
latex_elements = {"maketitle": latex_maketitle}



# SPDX-FileCopyrightText: BioPAL <biopal@esa.int>
# SPDX-License-Identifier: MIT

import os
import logging
import shutil
import numpy as np
from arepytools.io.productfolder import ProductFolder
from arepytools.io.metadata import DataSetInfo
from arepytools.timing.precisedatetime import PreciseDateTime
from arepytools.geometry.generalsarorbit import create_general_sar_orbit

from biopal.utility.utility_functions import (
    convert_rasterinfo_meters_to_seconds,
    convert_rasterinfo_seconds_to_meters,
)
from biopal.geometry import geometric_lib
from biopal.data_operations.data_operations import data_oversample
from biopal.utility.constants import (
    OVERSAMPLING_FACTOR,
    LIGHTSPEED,
)
from biopal.io.data_io import readBiomassHeader


def compute_and_oversample_geometry_auxiliaries(
    L1c_repository,
    aux_file_names,
    stack_id,
    acquisition_pf_names,
    master_id,
    enable_resampling,
    comp_ref_h=True,
    comp_dist=True,
    force_ellipsoid=False,
    sar_geometry_master=None,
):
    ecef_grid = None
    off_nadir_angle_rad = None
    kz = None
    reference_height = None
    R = None
    slope = None

    logging.info("Geometry library: loading auxiliary data: DEM...")
    pf_name = os.path.basename(aux_file_names.dem_folder)
    folder = os.path.dirname(aux_file_names.dem_folder)
    path_dem = os.path.join(folder, pf_name)

    try:
        pf_dem = ProductFolder(path_dem, "r")
    except Exception as e:
        logging.error("Geometry library:  error during auxiliary DEM reading: " + str(e), exc_info=True)
        raise

    ch_dem = pf_dem.get_channel(0)
    ri_dem = ch_dem.get_raster_info(0)

    dem_samples = ri_dem.samples
    lat_axis = (np.arange(0, ri_dem.lines) * ri_dem.lines_step + ri_dem.lines_start) * np.pi / 180
    lon_axis = (np.arange(0, dem_samples) * ri_dem.samples_step + ri_dem.samples_start) * np.pi / 180

    dem = pf_dem.read_data(0, [0, 0, ri_dem.lines, dem_samples])

    if force_ellipsoid:
        logging.info(
            "Geometry library: the ellipsoid altitude has been forced by input flag, only the slope over ellipsoid will be computed, after used to correcting geometry global / local reference."
        )
        dem = 0 * dem

    logging.info("...done\n")

    logging.info("Geometry library: loading orbits from stack metadata...")
    ch_list = []
    dsi_list = []
    swath_id_list = []
    swath_info_list = []

    # read master alone to absure it will be first in the list
    pf_master = ProductFolder(os.path.join(L1c_repository, master_id), "r")
    ch_master = pf_master.get_channel(0)
    ch_list.append(ch_master)
    ri_temp = ch_master.get_raster_info(0)

    (_, _, _, _, _, _, _, _, sensor_velocity,) = readBiomassHeader(
        ProductFolder(os.path.join(L1c_repository, master_id), "r"), 0
    )

    if ri_temp.samples_step_unit == "m":
        ri_master = convert_rasterinfo_meters_to_seconds(ri_temp, sensor_velocity)

        # ri_list.append( ri_master )
        num_samples = ri_temp.samples
        num_lines = ri_temp.lines
        lines_start = ri_temp.lines_start
        lines_start_unit = ri_temp.lines_start_unit
        lines_step = ri_temp.lines_step
        lines_step_unit = ri_temp.lines_step_unit
        samples_start = ri_temp.samples_start
        samples_start_unit = ri_temp.samples_start_unit
        samples_step = ri_temp.samples_step
        samples_step_unit = ri_temp.samples_step_unit

    else:
        ri_master = ri_temp

        ri_meters = convert_rasterinfo_seconds_to_meters(ri_temp, sensor_velocity)
        # ri_list.append( ri_master )
        num_samples = ri_meters.samples
        num_lines = ri_meters.lines
        lines_start = ri_meters.lines_start
        lines_start_unit = ri_meters.lines_start_unit
        lines_step = ri_meters.lines_step
        lines_step_unit = ri_meters.lines_step_unit
        samples_start = ri_meters.samples_start
        samples_start_unit = ri_meters.samples_start_unit
        samples_step = ri_meters.samples_step
        samples_step_unit = ri_meters.samples_step_unit

    sv_master = geometric_lib.cut_state_vectors(
        ch_master.get_state_vectors(0),
        ri_master.lines_start,
        ri_master.lines_start + ri_master.lines * ri_master.lines_step,
    )

    # Initiaize sar geometry
    flag_compute_sar_geometry = False
    if not sar_geometry_master:
        flag_compute_sar_geometry = True
        sar_geometry_master = geometric_lib.SARGeometry(ri_master, sv_master, lat_axis, lon_axis, dem)
        sar_geometry_master.compute_xyz()

    dsi_master = ch_master.get_dataset_info(0)
    dsi_list.append(dsi_master)

    swath_id_list.append(ch_master.get_swath_info(0).swath)
    swath_info_list.append(ch_master.get_swath_info(0))

    for acq_id in acquisition_pf_names:
        if not acq_id == master_id:
            pf_slave = ProductFolder(os.path.join(L1c_repository, acq_id), "r")
            ch_slave = pf_slave.get_channel(0)
            ch_list.append(ch_slave)

            ri_temp = ch_slave.get_raster_info(0)
            if ri_temp.samples_step_unit == "m":
                ri_slave = convert_rasterinfo_meters_to_seconds(ri_temp, sensor_velocity)
            else:
                ri_slave = ri_temp

            # ri_list. append( ri_slave )
            dsi_list.append(ch_slave.get_dataset_info(0))
            swath_id_list.append(ch_slave.get_swath_info(0).swath)
            swath_info_list.append(ch_slave.get_swath_info(0))

    logging.info("...done\n")

    logging.info("Geometry library: computing auxiliary data: slope...")

    if flag_compute_sar_geometry:
        sar_geometry_master.compute_terrain_slope()

    slope = sar_geometry_master.terrain_slope

    if enable_resampling:
        slope = data_oversample(slope, OVERSAMPLING_FACTOR, ri_master)[0]

    if force_ellipsoid:

        slope = slope.transpose()

        return ecef_grid, off_nadir_angle_rad, slope, kz, reference_height, R, sar_geometry_master

    else:
        pf_name = os.path.basename(aux_file_names.slope_file_names[stack_id])
        folder_name = os.path.dirname(aux_file_names.slope_file_names[stack_id])
        out_pf_name = os.path.join(folder_name, stack_id)

        if os.path.exists(out_pf_name):
            logging.warning("Overwriting of Auxiliary slope for stack " + stack_id)
            shutil.rmtree(out_pf_name)
        if not os.path.exists(folder_name):
            os.makedirs(folder_name)

        # slope = sar_geometry_master.terrain_slope
        pf = ProductFolder(out_pf_name, "w")

        # append a channel with the correct dimensions
        pf.append_channel(num_lines, num_samples, "FLOAT32", header_offset=0)

        # write the ECEF coordinate to the channel
        pf.write_data(0, slope)

        # prepare the metadata elements
        data_channel_obj = pf.get_channel(0)
        metadata_obj = data_channel_obj.metadata
        metadatachannel_obj = metadata_obj.get_metadata_channels(0)

        # Raster Info
        ri = metadatachannel_obj.get_element("RasterInfo")
        ri.set_lines_axis(lines_start, lines_start_unit, lines_step, lines_step_unit)
        ri.set_samples_axis(samples_start, samples_start_unit, samples_step, samples_step_unit)

        # DataSetInfo: to insert the X, Y or Z description
        dsi_list[0].description = "Auxiliary data: Slope [rad]"
        metadatachannel_obj.insert_element(dsi_list[0])

        pf.write_metadata(0)

        pf = None

    print("...done.\n")

    logging.info("Geometry library: computing auxiliary data: ECEFGRID...")

    ecef_grid = {
        "X": sar_geometry_master.x_sar_coords,
        "Y": sar_geometry_master.y_sar_coords,
        "Z": sar_geometry_master.z_sar_coords,
    }

    if enable_resampling:
        ecef_grid = data_oversample(ecef_grid, OVERSAMPLING_FACTOR, ri_master)[0]

    pf_name = os.path.basename(aux_file_names.ECEF_grid_file_names[stack_id])
    folder_name = os.path.dirname(aux_file_names.ECEF_grid_file_names[stack_id])
    out_pf_name = os.path.join(folder_name, stack_id)

    if os.path.exists(out_pf_name):
        logging.warning("Overwriting of Auxiliary ECEFGRID for stack " + stack_id)
        shutil.rmtree(out_pf_name)
    if not os.path.exists(folder_name):
        os.makedirs(folder_name)

    pf = ProductFolder(out_pf_name, "w")
    dsi_ecefgrid = DataSetInfo("STRIPMAP", 435000000)
    dsi_ecefgrid.description = ""
    dsi_ecefgrid.image_type = "SLC"
    dsi_ecefgrid.projection = "SLANT RANGE"
    dsi_ecefgrid.side_looking = "RIGHT"
    dsi_ecefgrid.sensor_name = "BIOMASS"
    dsi_ecefgrid.sense_date = PreciseDateTime()
    dsi_ecefgrid.acquisition_station = "NOT_AVAILABLE"
    dsi_ecefgrid.processing_center = "NOT_AVAILABLE"
    dsi_ecefgrid.processing_date = PreciseDateTime()
    dsi_ecefgrid.processing_software = "NOT_AVAILABLE"

    coord_names = ["X", "Y", "Z"]
    for channel_idx, coord_name in enumerate(coord_names):

        pf.append_channel(num_lines, num_samples, "FLOAT32", header_offset=0)
        pf.write_data(channel_idx, ecef_grid[coord_name])

        # prepare the metadata elements
        data_channel_obj = pf.get_channel(channel_idx)
        metadata_obj = data_channel_obj.metadata
        metadatachannel_obj = metadata_obj.get_metadata_channels(0)

        # Raster Info
        ri = metadatachannel_obj.get_element("RasterInfo")
        ri.set_lines_axis(lines_start, lines_start_unit, lines_step, lines_step_unit)
        ri.set_samples_axis(samples_start, samples_start_unit, samples_step, samples_step_unit)

        # DataSetInfo: to insert the X, Y or Z description
        dsi_ecefgrid.description = "Auxiliary data: " + coord_name + " ECEF GRID [m]"
        metadatachannel_obj.insert_element(dsi_ecefgrid)

        pf.write_metadata(channel_idx)

    pf = None
    logging.info("...done\n")

    targets_coords = np.vstack(
        (
            sar_geometry_master.x_sar_coords.flatten(),
            sar_geometry_master.y_sar_coords.flatten(),
            sar_geometry_master.z_sar_coords.flatten(),
        )
    )

    if comp_ref_h:
        logging.info("Geometry library: computing auxiliary data: reference height...")
        reference_height = sar_geometry_master.dem_sar

        if enable_resampling:
            reference_height = data_oversample(reference_height, OVERSAMPLING_FACTOR, ri_master)[0]

        pf_name = os.path.basename(aux_file_names.reference_height_file_names[stack_id])
        folder_name = os.path.dirname(aux_file_names.reference_height_file_names[stack_id])
        out_pf_name = os.path.join(folder_name, stack_id)

        if os.path.exists(out_pf_name):
            logging.warning("Overwriting of Auxiliary reference height for stack " + stack_id)
            shutil.rmtree(out_pf_name)
        if not os.path.exists(folder_name):
            os.makedirs(folder_name)

        # create the Product Folder in writing mode:
        pf = ProductFolder(out_pf_name, "w")
        # append a channel with the correct dimensions
        pf.append_channel(num_lines, num_samples, "FLOAT32", header_offset=0)
        # write the DTM to PF (reference height)
        pf.write_data(0, reference_height)

        # prepare the metadata elements
        data_channel_obj = pf.get_channel(0)
        metadata_obj = data_channel_obj.metadata
        metadatachannel_obj = metadata_obj.get_metadata_channels(0)

        # Raster Info
        ri = metadatachannel_obj.get_element("RasterInfo")
        ri.set_lines_axis(lines_start, lines_start_unit, lines_step, lines_step_unit)

        ri.set_samples_axis(samples_start, samples_start_unit, samples_step, samples_step_unit)

        # DataSetInfo
        dsi = DataSetInfo("STRIPMAP", 435000000)
        dsi.description = "Auxiliary data: Reference height (DTM)"
        dsi.image_type = "SLC"
        dsi.projection = "SLANT RANGE"
        dsi.side_looking = "RIGHT"
        dsi.sensor_name = "BIOMASS"
        dsi.sense_date = PreciseDateTime()
        dsi.acquisition_station = "NOT_AVAILABLE"
        dsi.processing_center = "NOT_AVAILABLE"
        dsi.processing_date = PreciseDateTime()
        dsi.processing_software = "NOT_AVAILABLE"

        metadatachannel_obj.insert_element(dsi)

        pf.write_metadata(0)
        pf = None

        print("...done.\n")

    if comp_dist:
        logging.info("Geometry library: computing auxiliary data: slant range distances...")

        pf_name = os.path.basename(aux_file_names.slant_range_distances_file_names[stack_id])
        folder_name = os.path.dirname(aux_file_names.slant_range_distances_file_names[stack_id])
        out_pf_name = os.path.join(folder_name, stack_id)

        if os.path.exists(out_pf_name):
            logging.warning("Overwriting of Auxiliary slant range distances for stack " + stack_id)
            shutil.rmtree(out_pf_name)
        if not os.path.exists(folder_name):
            os.makedirs(folder_name)

        pf = ProductFolder(out_pf_name, "w")
        R = {}
        for channel_idx, channel_curr in enumerate(ch_list):

            if channel_idx == 0:
                # distance between targets and master orbit
                R[swath_id_list[channel_idx]] = sar_geometry_master.compute_distance_master()
            else:

                # distance between targets and slave orbit
                ri_temp = channel_curr.get_raster_info(0)
                if ri_temp.samples_step_unit == "m":
                    ri_slave = convert_rasterinfo_meters_to_seconds(ri_temp, sensor_velocity)
                else:
                    ri_slave = ri_temp

                sv_slave = geometric_lib.cut_state_vectors(
                    channel_curr.get_state_vectors(0),
                    ri_slave.lines_start,
                    ri_slave.lines_start + ri_slave.lines * ri_slave.lines_step,
                )

                gso_slave = create_general_sar_orbit(sv_slave)

                _, range_points_master_slave = geometric_lib.perform_inverse_geocoding(gso_slave, targets_coords)

                distance_slave = range_points_master_slave * LIGHTSPEED / 2
                distance_slave.shape = sar_geometry_master.x_sar_coords.shape

                R[swath_id_list[channel_idx]] = distance_slave

            if enable_resampling:
                R[swath_id_list[channel_idx]] = data_oversample(
                    R[swath_id_list[channel_idx]], OVERSAMPLING_FACTOR, ri_master
                )[0]

            # append a channel with the correct dimensions
            pf.append_channel(num_lines, num_samples, "FLOAT32", header_offset=0)

            # write the R coordinate to the channel
            pf.write_data(channel_idx, R[swath_id_list[channel_idx]])

            # prepare the metadata elements
            data_channel_obj = pf.get_channel(channel_idx)
            metadata_obj = data_channel_obj.metadata
            metadatachannel_obj = metadata_obj.get_metadata_channels(0)

            # Raster Info
            ri = metadatachannel_obj.get_element("RasterInfo")
            ri.set_lines_axis(lines_start, lines_start_unit, lines_step, lines_step_unit)
            ri.set_samples_axis(samples_start, samples_start_unit, samples_step, samples_step_unit)

            # DataSetInfo: to insert the X, Y or Z description
            dsi_list[channel_idx].description = "Auxiliary data: slant range distances [m]"
            metadatachannel_obj.insert_element(dsi_list[channel_idx])

            metadatachannel_obj.insert_element(swath_info_list[channel_idx])

            pf.write_metadata(channel_idx)
        pf = None

        print("...done.\n")

    logging.info("Geometry library: computing auxiliary data: off nadir...")

    pf_name = os.path.basename(aux_file_names.off_nadir_angle_file_names[stack_id])
    folder_name = os.path.dirname(aux_file_names.off_nadir_angle_file_names[stack_id])
    out_pf_name = os.path.join(folder_name, stack_id)

    if os.path.exists(out_pf_name):
        logging.warning("Overwriting of Auxiliary off nadir for stack " + stack_id)
        shutil.rmtree(out_pf_name)
    if not os.path.exists(folder_name):
        os.makedirs(folder_name)

    pf = ProductFolder(out_pf_name, "w")
    off_nadir_angle_rad = {}
    for channel_idx, channel_curr in enumerate(ch_list):

        if channel_idx == 0:
            # master
            sensor_positions_master = sar_geometry_master.get_sensor_positions_master()
            off_nadir_angle_rad[
                swath_id_list[channel_idx]
            ] = sar_geometry_master.compute_off_nadir_angles_from_positions(
                sensor_positions_master, sensor_positions_master
            )
        else:
            # slaves
            ri_temp = channel_curr.get_raster_info(0)
            if ri_temp.samples_step_unit == "m":
                ri_slave = convert_rasterinfo_meters_to_seconds(ri_temp, sensor_velocity)
            else:
                ri_slave = ri_temp

            sv_slave = geometric_lib.cut_state_vectors(
                channel_curr.get_state_vectors(0),
                ri_slave.lines_start,
                ri_slave.lines_start + ri_slave.lines * ri_slave.lines_step,
            )

            gso_slave = create_general_sar_orbit(sv_slave)

            azimuth_points_master_slave, _ = geometric_lib.perform_inverse_geocoding(gso_slave, targets_coords)

            sensor_positions_slave = gso_slave.get_position(azimuth_points_master_slave)
            sensor_positions_slave.shape = (3, *sar_geometry_master.x_sar_coords.shape)
            off_nadir_angle_rad[
                swath_id_list[channel_idx]
            ] = sar_geometry_master.compute_off_nadir_angles_from_positions(
                sensor_positions_master, sensor_positions_slave
            )

        if enable_resampling:
            off_nadir_angle_rad[swath_id_list[channel_idx]] = data_oversample(
                off_nadir_angle_rad[swath_id_list[channel_idx]], OVERSAMPLING_FACTOR, ri_master
            )[0]

        # append a channel with the correct dimensions
        pf.append_channel(num_lines, num_samples, "FLOAT32", header_offset=0)

        # write the ECEF coordinate to the channel
        pf.write_data(channel_idx, off_nadir_angle_rad[swath_id_list[channel_idx]])

        # prepare the metadata elements
        data_channel_obj = pf.get_channel(channel_idx)
        metadata_obj = data_channel_obj.metadata
        metadatachannel_obj = metadata_obj.get_metadata_channels(0)

        # Raster Info
        ri = metadatachannel_obj.get_element("RasterInfo")
        ri.set_lines_axis(lines_start, lines_start_unit, lines_step, lines_step_unit)
        ri.set_samples_axis(samples_start, samples_start_unit, samples_step, samples_step_unit)

        # DataSetInfo: to insert the X, Y or Z description
        dsi_list[channel_idx].description = "Auxiliary data: Incidence Angle [rad]"
        metadatachannel_obj.insert_element(dsi_list[channel_idx])

        metadatachannel_obj.insert_element(swath_info_list[channel_idx])

        pf.write_metadata(channel_idx)

    pf = None
    print("...done.\n)")

    logging.info("Geometry library: computing auxiliary data: KZ (wavenumbers)...")

    pf_name = os.path.basename(aux_file_names.kz_file_names[stack_id])
    folder_name = os.path.dirname(aux_file_names.kz_file_names[stack_id])
    out_pf_name = os.path.join(folder_name, stack_id)

    if os.path.exists(out_pf_name):
        logging.warning("Overwriting of Auxiliary KZ (wavenumbers) for stack " + stack_id)
        shutil.rmtree(out_pf_name)
    if not os.path.exists(folder_name):
        os.makedirs(folder_name)

    pf = ProductFolder(out_pf_name, "w")
    kz = {}
    for channel_idx, channel_curr in enumerate(ch_list):

        if channel_idx == 0:
            # Master
            wavelength = LIGHTSPEED / ch_master.get_dataset_info(0).fc_hz

            kz[swath_id_list[channel_idx]] = np.zeros((ri_master.lines, ri_master.samples))
        else:
            kz[swath_id_list[channel_idx]] = geometric_lib.compute_vertical_wavenumber_angles(
                wavelength, off_nadir_angle_rad[swath_id_list[0]], off_nadir_angle_rad[swath_id_list[channel_idx]],
            )

        if enable_resampling:
            kz[swath_id_list[channel_idx]] = data_oversample(
                kz[swath_id_list[channel_idx]], OVERSAMPLING_FACTOR, ri_master
            )[0]

        # append a channel with the correct dimensions
        pf.append_channel(num_lines, num_samples, "FLOAT32", header_offset=0)

        # write the KZ  to the channel
        pf.write_data(channel_idx, kz[swath_id_list[channel_idx]])

        # prepare the metadata elements
        data_channel_obj = pf.get_channel(channel_idx)
        metadata_obj = data_channel_obj.metadata
        metadatachannel_obj = metadata_obj.get_metadata_channels(0)

        # Raster Info
        ri = metadatachannel_obj.get_element("RasterInfo")
        ri.set_lines_axis(lines_start, lines_start_unit, lines_step, lines_step_unit)
        ri.set_samples_axis(samples_start, samples_start_unit, samples_step, samples_step_unit)

        # DataSetInfo: to insert the X, Y or Z description
        dsi_list[channel_idx].description = "Auxiliary data: KZ (Interferometric wavenumber matrix)"
        metadatachannel_obj.insert_element(dsi_list[channel_idx])

        metadatachannel_obj.insert_element(swath_info_list[channel_idx])

        pf.write_metadata(channel_idx)

    pf = None
    logging.info("...done\n")

    # Transpose all the outputs as done in the case of auxiliaries directly read (and not computed), see "read_and_oversample_aux_data"
    if not ecef_grid is None:
        for coord_name in coord_names:
            ecef_grid[coord_name] = ecef_grid[coord_name].transpose()

    for swath_id in swath_id_list:
        if not off_nadir_angle_rad is None:
            off_nadir_angle_rad[swath_id] = off_nadir_angle_rad[swath_id].transpose()
        if not kz is None:
            kz[swath_id] = kz[swath_id].transpose()
        if not R is None:
            R[swath_id] = R[swath_id].transpose()

    if not slope is None:
        slope = slope.transpose()

    if not reference_height is None:
        reference_height = reference_height.transpose()

    return ecef_grid, off_nadir_angle_rad, slope, kz, reference_height, R, sar_geometry_master



# SPDX-FileCopyrightText: Aresys S.r.l. <info@aresys.it>
# SPDX-License-Identifier: MIT


class ext_geodata_mosaic:
    def __init__(self, latmin, latmax, lonmin, lonmax, database_dir, data_type, output_dir, geoid_dir=None):

        raise NotImplementedError

    def generate_csv_with_coordinates(self):
        raise NotImplementedError

    def corr_geoid(self, lat, lon):
        raise NotImplementedError

    def get_mosaic(self):
        raise NotImplementedError

    def save_mosaic(self, plot=False, forcePermission=True):
        raise NotImplementedError

    def save_biodempp_metadata(self):
        raise NotImplementedError

    def Run(self):
        raise NotImplementedError



# SPDX-FileCopyrightText: BioPAL <biopal@esa.int>
# SPDX-License-Identifier: MIT

import logging
import numpy as np
import scipy as sp
from shapely.geometry import MultiPoint
from scipy.interpolate import interp2d
import scipy.stats
from osgeo import gdal
from osgeo.gdalconst import GA_ReadOnly
from equi7grid.equi7grid import Equi7Grid
from biopal.utility.utility_functions import choose_equi7_sampling
from collections import namedtuple

# AGB look up tables containing paths boundaries and progressive stack indexing
# LookupTableAGB is the standard look up
LookupTableAGB = namedtuple("LookupTableAGB", ["paths", "boundaries", "progressive"])


def initialize_inversion_parameters(eq7_sampling, geographic_grid_sampling, geographic_boundaries, proc_confAGB):

    e7g_intermediate = Equi7Grid(eq7_sampling)

    # input pixel spacing (assumes last part of the path is the resolution
    # as is default in the simulation script)
    dx_in = eq7_sampling  # int(indir.split('_')[-1][:-1])

    # N0: upper left corner for output map (UTM 32S)
    sub_grid_string, x_upper_left, y_upper_left = e7g_intermediate.lonlat2xy(
        geographic_boundaries.lon_min, geographic_boundaries.lat_max
    )
    N0 = y_upper_left
    E0 = x_upper_left

    # lower right corner for output map (UTM 32S)
    _, x_lower_right, y_lower_right = e7g_intermediate.lonlat2xy(
        geographic_boundaries.lon_max, geographic_boundaries.lat_min
    )
    NN = y_lower_right
    EN = x_lower_right

    # ground pixel spacing for output map (m)
    if geographic_grid_sampling > (proc_confAGB.product_resolution / 2):
        geographic_grid_sampling = proc_confAGB.product_resolution / 2
        logging.warning(
            'Input "geographic_grid_sampling" cannot be greater than configuration "product_resolution"/2, setting to {}'.format(
                geographic_grid_sampling
            )
        )
    else:
        geographic_grid_sampling = geographic_grid_sampling

    geographic_grid_sampling = choose_equi7_sampling(proc_confAGB.product_resolution, geographic_grid_sampling)

    dE = np.abs(eq7_sampling)
    dN = -np.abs(eq7_sampling)

    # roi spacing for output map (m)
    dE_roi = np.abs(proc_confAGB.distance_sampling_area)
    dN_roi = -np.abs(proc_confAGB.distance_sampling_area)

    # define ROI size
    wE_roi = np.abs(proc_confAGB.product_resolution)  # dx_in # (cannot be finer than input pixel size)
    wN_roi = -np.abs(proc_confAGB.product_resolution)  # -dx_in # (cannot be finer than input pixel size)

    # parameter block spacing (m)
    dE_par = np.abs(proc_confAGB.distance_parameter_block)
    dN_par = -np.abs(proc_confAGB.distance_parameter_block)

    # parameter block width (m)
    wE_par = np.abs(
        proc_confAGB.parameter_block_size
    )  # dE_par*2 # 3/2 times spacing means that each tile shares 33.3% with its neighbour
    wN_par = -np.abs(
        proc_confAGB.parameter_block_size
    )  # dN_par*2 # 3/2 times spacing means that each tile shares 33.3% with its neighbour

    # number of tests to run (i.e., subsets of cals)
    N_tests = int(proc_confAGB.number_of_tests)

    # adjust ROI spacings to be intere multiples of output map ground pixel spacing
    dE_roi_old = dE_roi
    dE_roi = round(dE_roi / dE) * dE
    dN_roi = -dE_roi
    logging.info(
        "ROI ground pixel spacing (east and north) adjusted from {} to {} [m] in order to be an intere multiple of output map ground pixel spacing ".format(
            dE_roi_old, dE_roi
        )
    )

    # adjust PAR spacings to be intere multiples of ROI ground pixel spacing
    dE_par_old = dE_par
    dE_par = round(dE_par / dE_roi) * dE_roi
    dN_par = -dE_par
    logging.info(
        "PAR ground pixel spacing (east and north) adjusted from {} to {} [m] in order to be an intere multiple of ROI pixel spacing ".format(
            dE_par_old, dE_par
        )
    )

    _,
    # first_pixel_north,
    # first_pixel_east,
    # last_pixel_north,
    # last_pixel_east,
    # pixel_size_east,
    # pixel_size_north,
    # sample_spacing_east,
    # sample_spacing_north,
    # sample_size_east,
    # sample_size_north,
    # block_spacing_east,
    # block_spacing_north,
    # block_size_east,
    # block_size_north,
    # number_of_subsets,
    # geographic_grid_sampling,
    # sub_grid_string,

    return (
        dx_in,
        N0,
        E0,
        NN,
        EN,
        dE,
        dN,
        dE_roi,
        dN_roi,
        wE_roi,
        wN_roi,
        dE_par,
        dN_par,
        wE_par,
        wN_par,
        N_tests,
        geographic_grid_sampling,
        sub_grid_string,
    )


def compute_processing_blocs_order(lut_cal, EE_par_coordinates, wE_par, NN_par_coordinates, wN_par):

    # compute cal areas and find the block in which the largest calibration area is
    num_cals = lut_cal.shape[0]
    cal_areas = np.zeros(num_cals)
    for idx_cal in np.arange(num_cals):
        east_length = lut_cal[idx_cal, 1] - lut_cal[idx_cal, 0]
        north_length = lut_cal[idx_cal, 3] - lut_cal[idx_cal, 2]
        cal_areas[idx_cal] = east_length * north_length
    largest_cal_coords = lut_cal[np.argsort(cal_areas)[-1], :-1]

    intersection_areas = compute_intersection_area(
        EE_par_coordinates,
        EE_par_coordinates + wE_par,
        NN_par_coordinates + wN_par,
        NN_par_coordinates,
        largest_cal_coords[0],
        largest_cal_coords[1],
        largest_cal_coords[2],
        largest_cal_coords[3],
    )

    par_block_index_curr = np.argsort(intersection_areas)[-1]

    # sort blocks by distance from currentblock, extract only the tiles which cover the current polygon and the entire studied area
    block_order = np.argsort(
        np.sqrt(
            (EE_par_coordinates - EE_par_coordinates[par_block_index_curr]) ** 2
            + (NN_par_coordinates - NN_par_coordinates[par_block_index_curr]) ** 2
        )
    )  # tile_order

    return par_block_index_curr, block_order


# Simple table look-up, without interpolation or extrapolation, just exact matching
def tableLookupInt(x, y, xx):
    yy = -np.int32(np.ones(xx[:, 0].shape))
    for irow in np.arange(x.shape[0]):
        yy[(xx[:, 0] == x[irow, 0]) & (xx[:, 1] == x[irow, 1])] = y[irow]
    return yy


# forward and inverse transforms for parameters
def forward(x):
    return 10 * np.log10(x)


def inverse(x):
    return 10 ** (0.1 * x)


# function for regularizing indices
def regularizeIndices(a, b, off):
    uni = np.unique(np.concatenate((a, b)))
    lut = np.column_stack((uni, off + np.arange(len(uni))))
    f = sp.interpolate.interp1d(
        np.concatenate((-np.ones(1), lut[:, 0])), np.concatenate((-np.ones(1), lut[:, 1])), kind="nearest",
    )
    return np.int32(f(a)), np.int32(f(b)), np.int32(lut)


def tableLookupFloat(x, y, xx):
    yy = np.nan * np.ones(xx[:, 0].shape)
    for irow in np.arange(x.shape[0]):
        yy[(xx[:, 0] == x[irow, 0]) & (xx[:, 1] == x[irow, 1])] = y[irow]
    return yy


def check_intersection(
    A_east_min, A_east_max, A_north_min, A_north_max, B_east_min, B_east_max, B_north_min, B_north_max,
):

    if (
        (A_east_min >= A_east_max).any()
        or (A_north_min >= A_north_max).any()
        or (B_east_min >= B_east_max).any()
        or (B_north_min >= B_north_max).any()
    ):
        raise ValueError(
            'Wrong inputs, at least one of the "min" variable is greater than (or equal to) the "max" one.'
        )

    A_intersects_B = (
        (A_east_min <= B_east_max)
        & (A_east_max >= B_east_min)
        & (A_north_min <= B_north_max)
        & (A_north_max >= B_north_min)
    )

    return A_intersects_B


def compute_intersection_area(
    A_east_min_vec, A_east_max_vec, A_north_min_vec, A_north_max_vec, B_east_min, B_east_max, B_north_min, B_north_max,
):

    B_coords = tuple(
        zip([B_north_min, B_north_min, B_north_max, B_north_max], [B_east_min, B_east_max, B_east_min, B_east_max],)
    )
    B_polygon_obj = MultiPoint(B_coords).convex_hull

    num_areas = A_east_min_vec.shape[0]
    intersection_areas = np.zeros(num_areas)
    for area_idx in np.arange(num_areas):
        A_coords = tuple(
            zip(
                [
                    A_north_min_vec[area_idx],
                    A_north_min_vec[area_idx],
                    A_north_max_vec[area_idx],
                    A_north_max_vec[area_idx],
                ],
                [
                    A_east_min_vec[area_idx],
                    A_east_max_vec[area_idx],
                    A_east_min_vec[area_idx],
                    A_east_max_vec[area_idx],
                ],
            )
        )
        A_polygon_obj = MultiPoint(A_coords).convex_hull

        intersection_areas[area_idx] = B_polygon_obj.intersection(A_polygon_obj).area

    return intersection_areas


def interp2d_wrapper(data_path, band_index_to_read, east_out_axis, north_out_axis, fill_value):

    data_driver = gdal.Open(data_path, GA_ReadOnly)
    data_in = data_driver.GetRasterBand(band_index_to_read).ReadAsArray()
    geotransform = data_driver.GetGeoTransform()
    east_axis_in = geotransform[0] + geotransform[1] * np.arange(data_driver.RasterXSize)
    north_axis_in = geotransform[3] + geotransform[5] * np.arange(data_driver.RasterYSize)
    data_driver = None

    # interpolation function
    min_input = np.nanmin(data_in)
    nan_mask = np.isnan(data_in)
    data_in[nan_mask] = -9999

    interp2d_fun = interp2d(east_axis_in, north_axis_in[::-1], data_in[::-1, :], fill_value=fill_value)
    interp2d_fun_mask = interp2d(east_axis_in, north_axis_in[::-1], nan_mask[::-1, :], fill_value=float(0))

    # interpolate
    data_out = interp2d_fun(east_out_axis, north_out_axis[::-1])[::-1, :]
    mask_out = np.ceil(interp2d_fun_mask(east_out_axis, north_out_axis[::-1]))[::-1, :]

    mask_out = np.logical_and(mask_out, np.isnan(data_out))
    data_out[mask_out.astype(int)] = -9999

    data_out[data_out < min_input] = np.nan
    data_out[mask_out.astype(int)] = np.nan

    return data_out


def merge_agb_intermediate(a_data, b_data, method="nan_mean"):

    if method == "nan_mean":
        a_data = np.nanmean(np.dstack((a_data, b_data)), axis=2)

    elif method == "overlap_mean":
        # mean all the stacks
        a_data_is_present = np.logical_not(np.isnan(a_data))
        b_data_is_present = np.logical_not(np.isnan(b_data))

        # where both are present, perform the mean
        a_data[np.logical_and(b_data_is_present, a_data_is_present)] = np.mean(
            a_data[np.logical_and(b_data_is_present, a_data_is_present)],
            b_data[np.logical_and(b_data_is_present, a_data_is_present)],
        )

        # where only b is present perform the sum
        a_data[np.logical_and(b_data_is_present, np.logical_not(a_data_is_present))] = np.sum(
            a_data[np.logical_and(b_data_is_present, np.logical_not(a_data_is_present))],
            b_data[np.logical_and(b_data_is_present, np.logical_not(a_data_is_present))],
        )

    return a_data


def mean_on_rois(
    stack_data_interp, EE_pixel_mesh, NN_pixel_mesh, EE_roi_axis, dE_roi, NN_roi_axis, dN_roi, method,
):

    # Mean on ROIs
    number_of_rois = len(EE_roi_axis) * len(NN_roi_axis)
    data_roi_means_vec = np.zeros(number_of_rois)

    roi_counter = 0
    for north_roi_max in NN_roi_axis:
        for east_roi_min in EE_roi_axis:
            roi_counter = roi_counter + 1

            east_roi_max = east_roi_min + dE_roi
            north_roi_min = north_roi_max + dN_roi

            # here use < and > instead of <= and >= because the pixel meshes has been defined in the centre of the pixels
            EE_axis_valid = np.logical_and(EE_pixel_mesh > east_roi_min, EE_pixel_mesh < east_roi_max)
            NN_axis_valid = np.logical_and(NN_pixel_mesh > north_roi_min, NN_pixel_mesh < north_roi_max)
            curr_roi_indexes_pixels = np.logical_and(EE_axis_valid, NN_axis_valid)

            if method == "mean":
                data_roi_means_vec[roi_counter - 1] = np.mean(stack_data_interp[curr_roi_indexes_pixels])
            elif method == "nan_mean":
                data_roi_means_vec[roi_counter - 1] = np.nanmean(stack_data_interp[curr_roi_indexes_pixels])
            elif method == "median":
                data_roi_means_vec[roi_counter - 1] = np.median(stack_data_interp[curr_roi_indexes_pixels])
            elif method == "nan_median":
                data_roi_means_vec[roi_counter - 1] = np.nanmedian(stack_data_interp[curr_roi_indexes_pixels])
            elif method == "mode":
                data_roi_means_vec[roi_counter - 1] = sp.stats.mode(stack_data_interp[curr_roi_indexes_pixels])[0]
            elif method == "sum":
                data_roi_means_vec[roi_counter - 1] = np.sum(stack_data_interp[curr_roi_indexes_pixels])

    return data_roi_means_vec


def get_projection_from_path(file_name):

    driver = gdal.Open(file_name, GA_ReadOnly)
    projection = driver.GetProjection()

    return projection



# SPDX-FileCopyrightText: Aresys S.r.l. <info@aresys.it>
# SPDX-License-Identifier: MIT

"""
Utilities
---------
"""

import os
from functools import wraps
import numpy as np

_ERROR_CHECK_TYPE = "Wrong {} type: {} != {}"

_ERROR_CHECK_SIZE_OF_NUMPY_ARRAY = "Vector {} has wrong size: {} != {}"
_ERROR_CHECK_DTYPE_OF_NUMPY_ARRAY = "Vector {} has wrong dtype: {} != {}"
_ERROR_CHECK_SHAPE_OF_NUMPY_ARRAY = "Vector {} has wrong shape: {} != {}"
_ERROR_CHECK_NDIM_OF_NUMPY_ARRAY = "Vector {} has wrong ndim: {} != {}"
_ERROR_CHECK_FIRST_AXIS_SIZE_OF_NUMPY_ARRAY = "Vector {} has wrong first axis size: {} != {}"


def check_type(data, dtype, name="", throw_on_error=True) -> bool:
    """
    Checks that data is of type dtype, raises a RunTimeError

    :param data: input variable
    :param dtype: expected type
    :param name: name of the variable (for the error msg) [optional]
    :param throw_on_error: whether to raise a RunTimeError on check failure [optional, default true]
    """
    if not isinstance(data, dtype):
        if throw_on_error:
            raise RuntimeError(_ERROR_CHECK_TYPE.format(name, type(data), dtype))
        return False
    return True


def check_size_of_numpy_array(vector: np.ndarray, size, name="", throw_on_error=True) -> bool:
    """
    Checks the size of vector, raises a RunTimeError if it is not as expected.

    (It does not check if it is a numpy array)

    :param vector: should be a np.ndarray
    :param size: expected size
    :param name: name of the variable (for the error msg) [optional]
    :param throw_on_error: whether to raise a RunTimeError on check failure [optional, default true]
    """
    if size != vector.size:
        if throw_on_error:
            raise RuntimeError(_ERROR_CHECK_SIZE_OF_NUMPY_ARRAY.format(name, vector.size, size))
        return False
    return True


def check_dtype_of_numpy_array(vector: np.ndarray, dtype, name="", throw_on_error=True) -> bool:
    """
    Checks the data type (dtype) of vector, raises a RunTimeError if not as expected.

    (It does not check if it is a numpy array)

    :param vector: should be a np.ndarray
    :param dtype: expected type of the data inside
    :param name: name of the variable (for the error msg) [optional]
    :param throw_on_error: whether to raise a RunTimeError on check failure [optional, default true]
    """
    if dtype != vector.dtype:
        if throw_on_error:
            raise RuntimeError(_ERROR_CHECK_DTYPE_OF_NUMPY_ARRAY.format(name, vector.dtype, dtype))
        return False
    return True


def check_shape_of_numpy_array(vector: np.ndarray, shape, name="", throw_on_error=True) -> bool:
    """
    Checks the data type (dtype) of vector, raises a RunTimeError if not as expected.

    (It does not check if it is a numpy array)

    :param vector: should be a np.ndarray
    :param shape: expected shape of the data inside
    :param name: name of the variable (for the error msg) [optional]
    :param throw_on_error: whether to raise a RunTimeError on check failure [optional, default true]
    """
    if shape != vector.shape or len(shape) != len(vector.shape):
        if throw_on_error:
            raise RuntimeError(_ERROR_CHECK_SHAPE_OF_NUMPY_ARRAY.format(name, vector.shape, shape))
        return False
    return True


def check_ndim_of_numpy_array(vector: np.ndarray, ndim, name="", throw_on_error=True) -> bool:
    """
    Checks the num of dimensions (ndim) of vector, raises a RunTimeError if not as expected.

    (It does not check if it is a numpy array)

    :param vector: should be a np.ndarray
    :param ndim: expected ndim of the data
    :param name: name of the variable (for the error msg) [optional]
    :param throw_on_error: whether to raise a RunTimeError on check failure [optional, default true]
    """
    if ndim != vector.ndim:
        if throw_on_error:
            raise RuntimeError(_ERROR_CHECK_NDIM_OF_NUMPY_ARRAY.format(name, vector.ndim, ndim))
        return False
    return True


def check_first_axis_size_of_numpy_array(vector: np.ndarray, first_axis_size, name="", throw_on_error=True) -> bool:
    """
    Checks that the size of the first axis (first_axis_size) of vector is equal to first_axis_size,
     raises a RunTimeError if not as expected.

     Example: check first_axis_size=3 is ok for (3,) (3,1,1), but not for (1,3,1)

    (It does not check if it is a numpy array)

    :param vector: should be a np.ndarray
    :param first_axis_size: expected first_axis_size of the data
    :param name: name of the variable (for the error msg) [optional]
    :param throw_on_error: whether to raise a RunTimeError on check failure [optional, default true]
    """
    if first_axis_size != vector.shape[0]:
        if throw_on_error:
            raise RuntimeError(
                _ERROR_CHECK_FIRST_AXIS_SIZE_OF_NUMPY_ARRAY.format(name, vector.shape[0], first_axis_size))
        return False
    return True


def input_data_to_numpy_array_with_checks(data, dtype=None, size=None, shape=None, ndim=None, first_axis_size=None,
                                          name="") -> np.ndarray:
    """
    It takes a data and returns data converted to a vector, after some checks

    If data is a np.ndarray it checks dtype, size, shap, ndim (if provided) and returns the data as it is.
    If data is a list it converts it to a np.ndarray and then performs the same checks as stated above
    If data is a single object, the object is converted to a numpy array (following, shape and ndim if given
    and then the checks are performed.
    be (1, ) if ndim is not provided, (1,1,...) of length ndim if ndim was provided.

    :param data: the data
    :param dtype: expected type
    :param size: expected size
    :param ndim: expected ndim
    :param shape: expected shape
    :param first_axis_size: expected value for shape[0]
    :param name: name for error msg

    :return: return a numpy vector with dtype size as shape as expect, or it raises a RunTimeError
    """
    if isinstance(data, list):
        data_array = np.array(data)
    elif isinstance(data, np.ndarray):
        data_array = data
    else:
        if shape is not None:
            data_array = np.full(shape, data)
        else:
            if ndim is not None:
                shape_out = tuple(1 for _ in range(ndim))
            else:
                shape_out = (1,)
            data_array = np.full(shape_out, data)

    check_numpy_array(data_array, dtype=dtype, size=size, shape=shape, ndim=ndim, first_axis_size=first_axis_size,
                      name=name)
    return data_array


def check_numpy_array(data, dtype=None, size=None, shape=None, ndim=None, first_axis_size=None, name="",
                      throw_on_error=True) -> bool:
    """
    Verify that data is a numpy.ndarray and that it satisfies the specified constraints.
    If some constraints is not as expected an exception is raised.
    To disable this behavior and just get a boolean specify throw_on_error=False.
    Specify a name to customize the error messages.

    :param data: the data
    :param dtype: expected type
    :param size: expected size
    :param ndim: expected ndim
    :param first_axis_size: shape[0]
    :param shape: expected shape
    :param name: name for error msg
    :param throw_on_error: whether to raise or not an exception on check failure [optional, default true]
    """
    correctness = check_type(data, np.ndarray, name=name, throw_on_error=throw_on_error)
    if dtype is not None:
        correctness = correctness and check_dtype_of_numpy_array(data, dtype, name=name, throw_on_error=throw_on_error)
    if size is not None:
        correctness = correctness and check_size_of_numpy_array(data, size, name=name, throw_on_error=throw_on_error)
    if shape is not None:
        correctness = correctness and check_shape_of_numpy_array(data, shape, name=name, throw_on_error=throw_on_error)
    if ndim is not None:
        correctness = correctness and check_ndim_of_numpy_array(data, ndim, name=name, throw_on_error=throw_on_error)
    if first_axis_size is not None:
        correctness = correctness and check_first_axis_size_of_numpy_array(data, first_axis_size, name=name,
                                                                           throw_on_error=throw_on_error)
    return correctness


def _size_from_shape(shape: tuple) -> int:
    """
    From a tuple reprensenting a shape of a vector it retrives the size.

    For example (5, 1, 8) --> 40

    :param shape: the shape
    :return: the corresponding size
    """
    if not shape:
        return 0

    out = 1
    for k in shape:
        out *= k
    return out


def check_exists(path_to_file, tag=None):
    """
    Check if path exist, if not raise a runtime error
    :param path_to_file: the path to be checked
    :param tag: tag for error msg
    """
    if not os.path.exists(path_to_file):
        raise RuntimeError("{}: {} not found".format("Requested file" if tag is None else tag, path_to_file))


def check_file_exists(*positions):
    """
    Parametrized decorator.
    :param positions: position(s) of the arguments to be checked.
    Example:

    @check_file_exists(2, 4)
    def fun(a, b , path_in, c, path_in2, path_out):
        pass

    when calling fun the path_in and path_in2 will be validated
    the default arguments will be ignored
    """
    for pos in positions:
        if not isinstance(pos, int):
            raise RuntimeError("Decorator parameter should be one or more int")

    def decorator(fun):
        @wraps(fun)
        def decorated_fun(*args_fun):
            for position in positions:
                if position < len(args_fun):
                    check_exists(args_fun[position], "Argument in position {}".format(position))
            return fun(*args_fun)

        return decorated_fun

    return decorator



from pathlib import Path
import sys
import os
from matplotlib import pyplot as plt
import numpy as np

biopal_path = Path("C:\ARESYS_PROJ\BioPAL")

sys.path.append(str(biopal_path))
os.chdir(biopal_path)

from biopal.io.data_io import BiomassL1cRaster, BiomassL2Raster
from biopal.utility.plot import *

input_path = r"C:\bio\demo_input_file_run_AGB.xml"

from biopal.dataset_query.dataset_query import dataset_query

dataset_query_obj = dataset_query()
input_path_from_query = dataset_query_obj.run(input_path)

from biopal.io.xml_io import parse_input_file

input_params_obj = parse_input_file(input_path_from_query)
stack_based_processing_obj = input_params_obj.stack_based_processing
found_stack_ids = stack_based_processing_obj.stack_composition.keys()
print("SLC SAR images (slant range geometry) stacks and acquisitions found from the query")
for stack_id, acquisition_ids in stack_based_processing_obj.stack_composition.items():
    print("\n    Stack:")
    print("    ", stack_id)
    print("        Acquisitions:")
    for acq_id in acquisition_ids:
        print("        ", acq_id)
dataSet_path = input_params_obj.dataset_query.L1C_repository
print("\n The above stacks with the acquisitions can be found at following path:\n {}".format(dataSet_path))
# Print the path of the DTM projected in radar coordinates
reference_height_path = Path.home().joinpath(
    dataSet_path, stack_based_processing_obj.reference_height_file_names[stack_id]
)
print(
    "\n Full path of the reference height file valid for the above stack {}: \n {}".format(
        stack_id, reference_height_path
    )
)


#################### Plotting one of the SLC
slc_pf_path = Path.home().joinpath(dataSet_path, acq_id)

# initialize the SLC data object (its an L1C Raster)
SLC_obj = BiomassL1cRaster(slc_pf_path, channel_to_read=2)

# BioPAL offers different ways to plot a data with internal methods "plot_db", "plot_abs", "plot_angle", "plot":

# 1) Automatic plot
plot_db(SLC_obj)
plt.show()

# 2) Automatic plot with custom inputs:
plot_abs(
    SLC_obj, title="custom title 1 (ABS example)", clims=[0, 0.4], xlims=[8, 10], ylims=[6, 8],
)
plt.show()

# 3) Passing a pyplot axis to the automatic plot, than customize the axis as desired:
fig, ax = plt.subplots()
plot_angle(SLC_obj, ax=ax)
ax.set_xlabel("my xlabel")
ax.set_ylabel("my ylabel")
ax.set_title("custom title 2 (Angle example)")
ax.set_xlim([8, 10])
ax.set_ylim([6, 8])
fig.show()

# 4) Manual plot, from SLC_obj attributes:
plt.figure()
plt.imshow(
    10 * np.log10(np.abs(SLC_obj.data) ** 2),
    extent=[SLC_obj.x_axis[0], SLC_obj.x_axis[-1], SLC_obj.y_axis[-1], SLC_obj.y_axis[0],],
    cmap="gray",
)
plt.xlabel(SLC_obj.x_axis_description)
plt.ylabel(SLC_obj.y_axis_description)
plt.title("custom title 3 manual example")
plt.show()

# From here, into this Notebook automatic plot functions are used.


channel_to_read = 0  # Counter is zero based, there is only one channel
reference_height_obj = BiomassL1cRaster(reference_height_path, channel_to_read)
plot_db(reference_height_obj)
plt.show()

from biopal.agb.main_AGB import StackBasedProcessingAGB

configuration_file_path = str(Path.home().joinpath(biopal_path, "biopal", "conf", "Configuration_File.xml"))

# Initialize Stack Based Processing AGB APP
stack_based_processing_obj = StackBasedProcessingAGB(configuration_file_path)

# Run Stack Based Processing AGB APP
print("AGB stack-based processing APP started...")
(input_file_from_stack_based, configuration_file_updated,) = stack_based_processing_obj.run(input_path_from_query)

# Some of the computed output paths:
import os

input_params_obj = parse_input_file(input_file_from_stack_based)
output_folder = input_params_obj.output_specification.output_folder

npy_name = str(list(Path(Path.home().joinpath(output_folder, "Products", "breakpoints")).rglob("*.npy"))[0])
ground_canc_sr_path = Path.home().joinpath(output_folder, npy_name,)
print("\n Path of ground cancelled data in slant range geometry: \n {}".format(ground_canc_sr_path))

geocoded_dir = str(list(Path(Path.home().joinpath(output_folder, "Products", "temp", "geocoded")).rglob("GC_*"))[0])
sigma_tif_name = str(list(Path(geocoded_dir).rglob("sigma0_vh.tif"))[0])
ground_canc_gr_path = Path.home().joinpath(geocoded_dir, sigma_tif_name,)
print("\n  Path of ground cancelled VH data, geocoded: \n {}".format(ground_canc_gr_path))

theta_tif_name = str(list(Path(geocoded_dir).rglob("theta.tif"))[0])
theta_inc_path = Path.home().joinpath(geocoded_dir, theta_tif_name,)
print("\n  Path of incidence angle map, geocoded: \n {}".format(theta_inc_path))

# initialize the geocoded tif data object
# (BiomassL2Raster can be used both for intermediate gecoded and for output EQUI7 products)
ground_canc_obj = BiomassL2Raster(str(ground_canc_gr_path), band_to_read=1)
plot_db(ground_canc_obj)
plt.show()


theta_inc_obj = BiomassL2Raster(str(theta_inc_path), band_to_read=1)
theta_inc_figure = plot(theta_inc_obj)
plt.show()

from biopal.agb.main_AGB import CoreProcessingAGB

# Initialize Core Processing AGB APP
agb_processing_obj = CoreProcessingAGB(configuration_file_updated)

# Run Main APP #2: AGB Core Processing
print("AGB core-processing APP started (this will take some time, wait for ending message)...")
agb_processing_obj.run(input_file_from_stack_based)


input_params_obj = parse_input_file(input_file_from_stack_based)
output_folder = input_params_obj.output_specification.output_folder
tile_equi7_folder = list(Path.home().joinpath(Path(output_folder), "Products", "global_AGB").rglob("*"))[0]
tile_equi7_subfolder = list(Path(tile_equi7_folder).rglob("*"))[0]
final_estimation_path = list(Path(tile_equi7_subfolder).rglob("*.tif"))[0]
print("\n Path of the final AGB estimation product, in EQUI7 map geometry: \n {}".format(final_estimation_path))

reference_agb_folder = input_params_obj.stack_based_processing.reference_agb_folder
calibration_path = Path.home().joinpath(reference_agb_folder, "cal_05_no_errors.tif")
print("\n Path of the input calibration data used: \n {}".format(calibration_path))


lidar_agb_path = r"C:\ARESYS_PROJ\workingDir\biopal_data_V2_update\lope_lidar\equi7_50m\lidar_agb\EQUI7_AF050M\E045N048T3\lidar_AGB_AF050M_E045N048T3.tif"
if not Path(lidar_agb_path).exists():
    raise ValueError("Path of Lidar AGB - Reference map for validation should be set manually")

lidar_agb_obj = BiomassL2Raster(lidar_agb_path, band_to_read=1)
lidar_agb_figure = lidar_agb_obj.plot()
plt.show()

theta_inc_obj = BiomassL2Raster(lidar_agb_path, band_to_read=1)
plot(theta_inc_obj)
plt.show()


calibration_path = str(Path.home().joinpath(reference_agb_folder, "cal_05_no_errors.tif"))
band_to_read = 1
calibration_agb_obj = BiomassL2Raster(calibration_path, band_to_read)
calibration_agb_obj.plot(title="il mio titolo")
plt.show()

calibration_agb_obj = BiomassL2Raster(calibration_path, band_to_read)
plot(calibration_agb_obj)
plt.show()



from pathlib import Path
import sys
import os
from matplotlib import pyplot as plt
import numpy as np

biopal_path = Path("C:\ARESYS_PROJ\BioPAL")

sys.path.append(str(biopal_path))
os.chdir(biopal_path)

from biopal.io.data_io import BiomassL1cRaster, BiomassL2Raster
from biopal.utility.plot import *

_path = r"C:\ARESYS_PROJ\workingDir\biopal_data_V2_update\demo_lope_two\dataSet\GC_02_H_275.00_RGSW_00_RGSBSW_00_AZSW_00_BSL_00"
SLC_obj = BiomassL1cRaster(_path, channel_to_read=2)
plot_db(SLC_obj, title="SLC_obj")
plt.show()

_path = r"C:\bio\outNotebook\AGB\Products\temp\geocoded\GC_02_H_230.00_RGSW_00_RGSBSW_00_AZSW_00\theta.tif"
theta_obj = BiomassL2Raster(_path, intermediate_latlon_flag=True)
plot_rad2deg(theta_obj, title="theta_obj")
plt.show()

_path = r"C:\bio\outNotebook\AGB\Products\temp\geocoded\GC_02_H_230.00_RGSW_00_RGSBSW_00_AZSW_00\theta.tif"
theta_obj = BiomassL2Raster(_path)
plot_rad2deg(theta_obj, title="theta_obj wrong labels")
plt.show()

_path = r"C:\bio\outNotebook\AGB\Products\temp\geocoded\GC_02_H_230.00_RGSW_00_RGSBSW_00_AZSW_00\sigma0_hh.tif"
sigma_obj = BiomassL2Raster(_path, band_to_read=1, intermediate_latlon_flag=True)
plot(sigma_obj, title="sigma_obj")
plt.show()


_path = (
    r"C:\ARESYS_PROJ\workingDir\biopal_data_V2_update\demo_lope_two\auxiliary_data_pf\ReferenceAGB\cal_05_no_errors.tif"
)
cal_agb_obj = BiomassL2Raster(_path, band_to_read=1)
plot(cal_agb_obj)
plt.show()

_path = r"C:\bio\outNotebook\AGB\Products\global_AGB\AF050M\E045N048T3\agb_1_est_db_backtransf_.tif"
agb_obj = BiomassL2Raster(_path)
plot(agb_obj, xlims=[4512, 4523], ylims=[5033, 5043])
plt.show()
